import os
import logging
import pathlib
import tempfile
import importlib

import pytest

import astropy
import astropy.config.paths
import astropy.io.fits
from astropy.utils import iers

from sunpy.data.test import get_test_data_filenames, get_test_filepath, write_image_file_from_header_file
from sunpy.util import SunpyUserWarning

# Force MPL to use non-gui backends for testing.
try:
    import matplotlib
    import matplotlib.pyplot as plt

    HAVE_MATPLOTLIB = True
    matplotlib.use('Agg')
except ImportError:
    HAVE_MATPLOTLIB = False

# Don't actually import pytest_remotedata because that can do things to the
# entrypoints code in pytest.
remotedata_spec = importlib.util.find_spec("pytest_remotedata")
HAVE_REMOTEDATA = remotedata_spec is not None
# Do not collect the sample data file because this would download the sample data.
collect_ignore = ["data/sample.py"]
console_logger = logging.getLogger()
console_logger.setLevel('INFO')


@pytest.fixture
def jsoc_test_email():
    # See https://github.com/sunpy/sunpy/wiki/Home:-JSOC
    return "jsoc@sunpy.org"


@pytest.fixture(scope='session', autouse=True)
def no_download_iers(request):
    # Don't try and download IERS during tests
    # See https://github.com/astropy/astropy/issues/12998 for issue that this
    # sidesteps
    old_value = iers.conf.auto_download
    iers.conf.auto_download = False
    yield
    iers.conf.auto_download = old_value


@pytest.fixture(scope='session', autouse=True)
def tmp_config_dir(request):
    """
    Globally set the default config for all tests.
    """
    tmpdir = tempfile.TemporaryDirectory()

    os.environ["SUNPY_CONFIGDIR"] = str(tmpdir.name)
    astropy.config.paths.set_temp_config._temp_path = pathlib.Path(tmpdir.name)
    astropy.config.paths.set_temp_cache._temp_path = pathlib.Path(tmpdir.name)

    yield

    del os.environ["SUNPY_CONFIGDIR"]
    tmpdir.cleanup()
    astropy.config.paths.set_temp_config._temp_path = None
    astropy.config.paths.set_temp_cache._temp_path = None


@pytest.fixture()
def sunpy_cache(mocker, tmp_path):
    """
    Provide a way to add local files to the cache. This can be useful when mocking
    remote requests.
    """
    from types import MethodType

    from sunpy.data.data_manager.cache import Cache
    from sunpy.data.data_manager.downloader import ParfiveDownloader
    from sunpy.data.data_manager.storage import InMemStorage
    cache = Cache(
        ParfiveDownloader(),
        InMemStorage(),
        tmp_path,
        None
    )

    def add(self, url, path):
        self._storage.store({
            'url': url,
            'file_path': path,
            'file_hash': 'none',  # hash doesn't matter
        })
    cache.add = MethodType(add, cache)

    def func(mocked):
        mocker.patch(mocked, cache)
        return cache
    return func


@pytest.fixture()
def undo_config_dir_patch():
    """
    Provide a way for certain tests to not have the config dir.
    """
    oridir = os.environ["SUNPY_CONFIGDIR"]
    del os.environ["SUNPY_CONFIGDIR"]
    yield
    os.environ["SUNPY_CONFIGDIR"] = oridir


@pytest.fixture(scope='session', autouse=True)
def sunpy_test_run(request):
    """
    Add a environmental keyword to allow us to check and execute
    specific code within a pytest run (e.g., update headers for the online tests).
    """
    os.environ["SUNPY_PYTEST_RUN"] = "True"
    yield
    del os.environ["SUNPY_PYTEST_RUN"]


@pytest.fixture(scope='session', autouse=True)
def hide_parfive_progress(request):
    """
    Set the PARFIVE_HIDE_PROGRESS to hide the parfive progress bar in tests.
    """
    os.environ["PARFIVE_HIDE_PROGRESS"] = "True"
    yield
    del os.environ["PARFIVE_HIDE_PROGRESS"]


@pytest.fixture(scope='session', autouse=True)
def tmp_dl_dir(request):
    """
    Globally set the default download directory for the test run to a tmp dir.
    """
    with tempfile.TemporaryDirectory() as tmpdir:
        os.environ["SUNPY_DOWNLOADDIR"] = tmpdir
        yield tmpdir
        del os.environ["SUNPY_DOWNLOADDIR"]


@pytest.fixture(scope='session', autouse=True)
def set_columns(request):
    orig_columns = os.environ.get("COLUMNS", None)
    os.environ['COLUMNS'] = '180'
    yield
    if orig_columns is not None:
        os.environ['COLUMNS'] = orig_columns


@pytest.fixture()
def undo_download_dir_patch():
    """
    Provide a way for certain tests to not have tmp download dir.
    """
    oridir = os.environ["SUNPY_DOWNLOADDIR"]
    del os.environ["SUNPY_DOWNLOADDIR"]
    yield
    os.environ["SUNPY_DOWNLOADDIR"] = oridir


def pytest_runtest_setup(item):
    """
    pytest hook to skip all tests that have the mark 'remotedata' if the
    pytest_remotedata plugin is not installed.
    """
    if isinstance(item, pytest.Function):
        if 'remote_data' in item.keywords and not HAVE_REMOTEDATA:
            pytest.skip("skipping remotedata tests as pytest-remotedata is not installed")

    # Confirm that the pyplot figure stack is empty before the test
    if HAVE_MATPLOTLIB and plt.get_fignums():
        raise SunpyUserWarning(f"There are stale pyplot figures prior to running {item.name}")


def pytest_runtest_teardown(item):
    # Clear the pyplot figure stack if it is not empty after the test
    # You can see these log messages by passing "-o log_cli=true" to pytest on the command line
    if HAVE_MATPLOTLIB and plt.get_fignums():
        console_logger.info(f"Removing {len(plt.get_fignums())} pyplot figure(s) "
                            f"left open by {item.name}")
        plt.close('all')


@pytest.fixture(scope="session")
def aia171_test_map():
    from sunpy.map import Map
    return Map(get_test_filepath('aia_171_level1.fits'))


@pytest.fixture(scope="session")
def eit_fits_directory(tmp_path_factory):
    # Create a temporary directory of dummy EIT FITS files
    # from the header data. This directory is then used to
    # test directory and glob patterns for the map factory
    eit_dir = tmp_path_factory.mktemp('EIT')
    eit_header_files = [f for f in get_test_data_filenames()
                        if f.parents[0].relative_to(f.parents[1]).name == 'EIT_header'
                        and f.suffix == '.header']
    for f in eit_header_files:
        _ = write_image_file_from_header_file(f, eit_dir)
    return pathlib.Path(eit_dir)



import numpy as np
import pytest

from sunpy.io import ana
from sunpy.tests.helpers import skip_ana

pytestmark = [
    skip_ana,
    pytest.mark.filterwarnings("ignore::sunpy.util.exceptions.SunpyDeprecationWarning"),
]


@pytest.fixture(scope="module")
def img_size():
    return (456, 345)


@pytest.fixture(scope="module")
def img_src(img_size):
    img = np.arange(np.prod(img_size))
    img.shape = img_size
    return img


@pytest.fixture(scope="module")
def img_i8(img_src):
    arr = img_src * 2**8 / img_src.max()
    return arr.astype(np.int8)


@pytest.fixture(scope="module")
def img_i16(img_src):
    arr = img_src * 2**16 / img_src.max()
    return arr.astype(np.int16)


@pytest.fixture(scope="module")
def img_f32(img_src):
    arr = img_src * 1.0 / img_src.max()
    return arr.astype(np.float32)


def test_roundtrip_int8_compressed(img_i8, tmp_path):
    p = tmp_path / "i8_compressed.ana"
    ana.write(str(p), img_i8, comments="testcase", compress=True)
    (data, _), = ana.read(str(p))
    np.testing.assert_array_equal(data, img_i8)


def test_roundtrip_int8_uncompressed(img_i8, tmp_path):
    p = tmp_path / "i8_uncompressed.ana"
    ana.write(str(p), img_i8, comments="testcase", compress=False)
    (data, _), = ana.read(str(p))
    np.testing.assert_array_equal(data, img_i8)


def test_roundtrip_int16_compressed(img_i16, tmp_path):
    p = tmp_path / "i16_compressed.ana"
    ana.write(str(p), img_i16, comments="testcase", compress=True)
    (data, _), = ana.read(str(p))
    np.testing.assert_array_equal(data, img_i16)


def test_roundtrip_int16_uncompressed(img_i16, tmp_path):
    p = tmp_path / "i16_uncompressed.ana"
    ana.write(str(p), img_i16, comments="testcase", compress=False)
    (data, _), = ana.read(str(p))
    np.testing.assert_array_equal(data, img_i16)


def test_roundtrip_float32_uncompressed(img_f32, tmp_path):
    p = tmp_path / "f32_uncompressed.ana"
    ana.write(str(p), img_f32, comments="testcase", compress=False)
    (data, _), = ana.read(str(p))
    np.testing.assert_array_equal(data, img_f32)


def test_roundtrip_float32_compressed(img_f32, tmp_path):
    p = tmp_path / "f32_compressed.ana"
    ana.write(str(p), img_f32, comments="testcase", compress=True)
    (data, _), = ana.read(str(p))
    np.testing.assert_array_equal(data, img_f32)


def test_memmap_read_returns_copy(img_f32, tmp_path):
    p = tmp_path / "f32_memmap.ana"
    ana.write(str(p), img_f32, comments="testcase", compress=False)
    # Memmap mode should still return an array that owns its data (no view)
    (data_memmap, _), = ana.read(str(p), memmap=True)
    assert data_memmap.base is None
    (data, _), = ana.read(str(p), memmap=False)
    assert data.base is None
    np.testing.assert_array_equal(data_memmap, data)



import os
import re
import glob
import pathlib
import collections
import urllib.request

import fsspec

HDPair = collections.namedtuple("HDPair", ["data", "header"])


def expand_fsspec_open_file(open_file):
    """
    Expand an OpenFile object to apply a glob if it is a directory.
    """
    fs = open_file.fs
    path = open_file.path

    if not fs.exists(path):
        raise FileNotFoundError(f"{open_file.full_name} does not exist.")

    if fs.isdir(path):
        return fsspec.open_files([fs.unstrip_protocol(f) for f in sorted(fs.glob(path + "/*"))])

    return [open_file]


def parse_path(path, f, **kwargs):
    """
    Read in a series of files at *path* using the function *f*.

    Parameters
    ----------
    path : pathlib.Path
    f : callable
        Must return a list of read-in data.
    kwargs :
        Additional keyword arguments are handed to ``f``.

    Returns
    -------
    list
        List of files read in by ``f``.
    """
    if not isinstance(path, os.PathLike):
        raise ValueError("path must be a pathlib.Path object")
    path = path.expanduser()
    if is_file(path):
        return f(path, **kwargs)
    elif is_dir(path):
        read_files = []
        for afile in sorted(path.glob("*")):
            read_files += f(afile, **kwargs)
        return read_files
    elif glob.glob(str(path)):
        read_files = []
        for afile in sorted(glob.glob(str(path))):
            afile = pathlib.Path(afile)
            read_files += f(afile, **kwargs)
        return read_files
    else:
        raise ValueError(f"Did not find any files at {path}")


# In python<3.8 paths with un-representable chars (ie. '*' on windows)
# raise an error, so make our own version that returns False instead of
# erroring. These can be removed when we support python >= 3.8
# https://docs.python.org/3/library/pathlib.html#methods
def is_file(path):
    try:
        return path.is_file()
    except Exception:
        return False


def is_dir(path):
    try:
        return path.is_dir()
    except Exception:
        return False


def possibly_a_path(obj):
    """
    Check if ``obj`` can be coerced into a pathlib.Path object.
    Does *not* check if the path exists.
    """
    try:
        pathlib.Path(obj)
        return True
    except Exception:
        return False


def is_url(obj):
    """
    Check if the given object is a valid URL.

    Parameters
    ----------
    obj : str
        The object to check.

    Returns
    -------
    bool
        True if the object is a valid URL, False otherwise.
    """
    try:
        urllib.request.urlopen(obj)
        return True
    except Exception:
        return False


def is_uri(obj):
    """
    Check if the given object is a valid non-http(s) URI.

    Parameters
    ----------
    obj : str
        The object to check.

    Returns
    -------
    bool
        True if the object is a valid URI, False otherwise.
    """
    try:
        _RFC3896_ = re.compile(r"^[A-Za-z][A-Za-z0-9+\-+.]*://")
        return bool(_RFC3896_.match(obj)) and not obj.startswith(("http://", "https://"))
    except Exception:
        return False


def string_is_float(s):
    try:
        float(s)
        return True
    except ValueError:
        return False



"""
========================
Creating a Composite map
========================

How to create a composite map and use it to overplot two maps to
compare features.
"""
import matplotlib.pyplot as plt

import astropy.units as u
from astropy.coordinates import SkyCoord

import sunpy.data.sample
import sunpy.map

###############################################################################
# We start with the sample data. HMI shows the line-of-sight magnetic field at
# the photosphere while AIA 171 images show the resulting magnetic fields
# filled with hot plasma above, in the corona. We want to see what coronal
# features overlap with regions of strong line-of-sight magnetic fields.

aia_map = sunpy.map.Map(sunpy.data.sample.AIA_171_IMAGE)
hmi_map = sunpy.map.Map(sunpy.data.sample.HMI_LOS_IMAGE)

bottom_left = [0, 0] * u.arcsec
top_right = [800, 800] * u.arcsec
aia_smap = aia_map.submap(SkyCoord(*bottom_left, frame=aia_map.coordinate_frame),
                          top_right=SkyCoord(*top_right, frame=aia_map.coordinate_frame))
hmi_smap = hmi_map.submap(SkyCoord(*bottom_left, frame=hmi_map.coordinate_frame),
                          top_right=SkyCoord(*top_right, frame=hmi_map.coordinate_frame))

##############################################################################
# Let's create a `~sunpy.map.CompositeMap` which includes both maps.

comp_map = sunpy.map.Map(aia_smap, hmi_smap, composite=True)

# Let's set the contours of the HMI map, the second image in our composite map
# (therefore the index is 1), from a few hundred to a thousand Gauss which
# is the typical field associated with umbral regions of active regions.
levels = [-1000, -500, -250, 250, 500, 1000] * u.G
comp_map.set_levels(index=1, levels=levels)

##############################################################################
# Now let us look at the result. Notice that we can see the coronal structures
# present on the AIA image and how they correspond to the line of sight
# magnetic field.

fig = plt.figure()
ax = fig.add_subplot(projection=comp_map.get_map(0))
comp_map.plot(axes=ax)
plt.show()



"""
==============================================
Auto-Aligning AIA and HMI Data During Plotting
==============================================

This example shows how a map is autoaligned when it is plotted on a different
reference frame.

See :ref:`sphx_glr_generated_gallery_map_transformations_reprojection_align_aia_hmi.py`
for an alternate approach to image alignment, where one of the maps is modified
prior to plotting, and thus is available for purposes other than plotting.
"""
import matplotlib.pyplot as plt

import astropy.units as u

import sunpy.data.sample
import sunpy.map

######################################################################
# We use the AIA image and HMI image from the sample data. For the
# HMI map, we use the special HMI color map, which expects the plotted
# range to be -1500 to 1500.

map_aia = sunpy.map.Map(sunpy.data.sample.AIA_171_IMAGE)
map_hmi = sunpy.map.Map(sunpy.data.sample.HMI_LOS_IMAGE)
map_hmi.plot_settings['cmap'] = "hmimag"
map_hmi.plot_settings['norm'] = plt.Normalize(-1500, 1500)

######################################################################
# Plot both images side by side. Note by the tick labels that the HMI
# image is oriented "upside down" relative to the AIA image.

fig = plt.figure(figsize=(12, 5))
ax1 = fig.add_subplot(121, projection=map_aia)
map_aia.plot(axes=ax1, clip_interval=(1, 99.9)*u.percent)
ax2 = fig.add_subplot(122, projection=map_hmi)
map_hmi.plot(axes=ax2)

######################################################################
# Now let us intentionally set the ``projection`` for the right panel
# to be ``map_aia`` instead of ``map_hmi``. This time, plotting the
# HMI image onto axes defined by the AIA reference frame will trigger
# "autoalignment" functionality where each map pixel is individually
# drawn. The HMI image now has the same orientation as the AIA image.
#
# Note that off-disk HMI data are not retained by default because an
# additional assumption is required to define the location of the HMI
# emission in 3D space. We can use `~sunpy.coordinates.SphericalScreen`
# to retain the off-disk HMI data. See
# :ref:`sphx_glr_generated_gallery_map_transformations_reprojection_spherical_screen.py`
# for more reference.

fig = plt.figure(figsize=(12, 5))
ax1 = fig.add_subplot(121, projection=map_aia)
map_aia.plot(axes=ax1, clip_interval=(1, 99.9)*u.percent)
ax2 = fig.add_subplot(122, projection=map_aia)
map_hmi.plot(axes=ax2, title='HMI image in AIA reference frame')

######################################################################
# We can directly plot them over one another, by setting the
# transparency of the HMI plot.

fig = plt.figure()
ax1 = fig.add_subplot(projection=map_aia)
map_aia.plot(axes=ax1, clip_interval=(1, 99.9)*u.percent)
map_hmi.plot(axes=ax1, alpha=0.5)
ax1.set_title('HMI overlaid on AIA')

plt.show()

# sphinx_gallery_thumbnail_number = 2



"""
Allow representation of queries as logic expressions. This module makes
sure that attributes that are combined using the two logic operations AND (&)
and OR (|) always are in disjunctive normal form, that is, there are only two
levels ­- the first being disjunction and the second being conjunction. In other
words, every combinations of attributes looks like this:
(a AND b AND c) OR (d AND e).

Walkers are used to traverse the tree that results from combining attributes.
They are implemented using `functools.singledispatch` modified to dispatch on the second argument to the function.

Please note that & is evaluated first, so A & B | C is equivalent to
(A & B) | C.
"""
import re
import string
import inspect
import keyword
import textwrap
from textwrap import dedent
from collections import namedtuple, defaultdict

import numpy as np

from astropy.table import Table

from sunpy.extern import inflect
from sunpy.util.functools import seconddispatch
from sunpy.util.util import get_width

_ATTR_TUPLE = namedtuple("attr", "name client name_long desc")
# Matches any number.
NUMBER_REGEX = re.compile(r"^(\d+$|\d(?:\.\d+)?)")

__all__ = ['AttrMeta', 'Attr', 'DataAttr', 'DummyAttr', 'SimpleAttr', 'Range', 'AttrAnd', 'AttrOr',
           'ValueAttr', 'and_', 'or_', 'AttrWalker', 'AttrComparison', 'ComparisonParamAttrWrapper']


def make_tuple():
    return _ATTR_TUPLE([], [], [], [])


def _create_table(attr):
    """
    Create a table from the given attribute registry.

    Parameters
    ----------
    attr : object
        An object that contains an attribute registry.

    Returns
    -------
    astropy.table.Table
        A table with columns "Attribute Name", "Client", "Full Name", and "Description".
    """
    attrs = attr._attr_registry[attr]
    # Only sort the attrs if any have been registered
    sorted_attrs = _ATTR_TUPLE(*zip(*sorted(zip(*attrs)))) if attrs.name else make_tuple()
    *other_row_data, descs = sorted_attrs
    descs = [(dsc[:77] + '...') if len(dsc) > 80 else dsc for dsc in descs]
    table = Table(names=["Attribute Name", "Client", "Full Name", "Description"],
                  dtype=["U80", "U80", "U80", "U80"],
                  data=[*other_row_data, descs])
    return table


def _print_attrs(attr, html=False):
    """
    Given a Attr class will print out each registered attribute.

    Parameters
    ----------
    attr : `sunpy.net.attr.Attr`
        The attr class/type to print for.
    html : bool
        Will return a html table instead.

    Returns
    -------
    `str`
        String with the registered attributes.
    """
    table = _create_table(attr)

    class_name = f"{(attr.__module__ + '.') or ''}{attr.__name__}"
    lines = [class_name]
    # If the attr lacks a __doc__ this will error and prevent this from returning anything.
    try:
        lines.append(dedent(attr.__doc__.partition("\n\n")[0]) + "\n")
    except AttributeError:
        pass

    format_line = "<p>{}</p>" if html else "{}"
    width = -1 if html else get_width()

    lines = [*[format_line.format(line) for line in lines],
             *table.pformat(show_dtype=False, max_width=width, align="<", html=html)]
    return '\n'.join(lines)


class AttrMeta(type):
    """
    We want to enable discovery, by tab completion, of values for all subclasses of Attr.
    So have to create a metaclass that overloads the methods that Python uses, so that they work on the classes.
    This would allow that ``attrs.Instrument`` to be able to tab complete to ``attrs.Instrument.aia``.
    """

    # The aim is to register Attrs as a namedtuple of lists
    # So we define the namedtuple outside of AttrMeta, see above.
    _attr_registry = defaultdict(make_tuple)

    def __getattr__(self, item):
        """
        Our method for Attrs is to register using the attribute type (i.e. Instrument) as keys
        in a dictionary. ``_attr_registry`` is a dictionary with the keys being subclasses of Attr
        and the value being the namedtuple of lists.

        As a result we index `_attr_registry` with `[self]` which will be the `type`
        of the `Attr` class to access the dictionary. This will return the namedtuple
        that has three attributes: `name`, `name_long` and `desc`.
        Each of which are a list. `name` will be the attribute name, `name_long` is
        the original name passed in and `desc` the description of the object.
        """
        # Get the relevant entries.
        registry = self._attr_registry[self]
        # All the attribute names under that type(Attr)
        names = registry.name
        try:
            # We return Attr(name_long) to create the Attr requested.
            return self(registry.name_long[names.index(item)])
        except ValueError:
            raise AttributeError(f'This attribute, {item} is not defined, please register it.')

    def __dir__(self):
        """
        To tab complete in Python we need to add to the `__dir__()` return.
        So we add all the registered values for this subclass of Attr to the output.
        """
        custom_attrs = set(self._attr_registry[self].name)
        # "all" can be registered as a documentation helper, but isn't a valid attr
        custom_attrs.discard("all")
        return super().__dir__() + list(custom_attrs)

    def __repr__(self):
        """
        Returns the normal repr plus the pretty attr __str__.
        """
        return f"{type.__repr__(self)}\n{str(self)}"

    def __str__(self):
        """
        This enables the "pretty" printing of Attrs.
        """
        return _print_attrs(self)

    def _repr_html_(self):
        """
        This enables the "pretty" printing of Attrs with html.
        """
        return _print_attrs(self, html=True)


    def show_in_notebook(self, **kwargs):
        """
        Display the attrs tables as interactive grids in a Jupyter Notebook.

        This function utilizes the ``itables`` library to render tables as interactive grids.

        .. note::
            This function requires the optional dependency ``itables``.
            Ensure it is installed before calling this method.

        Parameters
        ----------
        **kwargs : dict, optional
            Additional keyword arguments to customize the ``itables.show`` function.

        """
        try:
            from itables import show
        except ImportError:
            raise ImportError(
                "`itables` is required to display tables. "
                "Install itables using `pip install itables` or `conda install -c conda-forge itables`."
            )
        table = _create_table(self)
        show(table.to_pandas(), **kwargs)


class Attr(metaclass=AttrMeta):
    """This is the base for all attributes."""

    def __and__(self, other):
        if isinstance(other, AttrOr):
            return AttrOr([elem & self for elem in other.attrs])
        if self.collides(other):
            return NotImplemented
        if isinstance(other, AttrAnd):
            return AttrAnd([self, *other.attrs])
        return AttrAnd([self, other])

    def __hash__(self):
        return hash(frozenset(vars(self).items()))

    def __or__(self, other):
        # Optimization.
        if self == other:
            return self
        if isinstance(other, AttrOr):
            return AttrOr([self, *other.attrs])
        return AttrOr([self, other])

    def collides(self, other):
        raise NotImplementedError

    def __eq__(self, other):
        if not isinstance(other, Attr):
            return False
        return vars(self) == vars(other)

    @classmethod
    def update_values(cls, adict):
        """
        Clients will use this method to register their values for subclasses of `~sunpy.net.attr.Attr`.

        The input has to be a dictionary, with each key being an instance of a client.
        The value for each client has to be a dictionary with each key being a subclass of Attr.
        The value for each Attr key should be a list of tuples with each tuple of the form
        ``(Name, Description)``.
        If you do not want to add a description, you can put `None` or an empty string.
        We sanitize the name you provide by removing all special characters and making it all lower case.
        If it still invalid we will append to the start of the name to make it a valid attribute name.
        This is an ugly workaround, so please try to pick a valid attribute name.

        **It is recommended that clients register values for all attrs used by their `_can_handle_query` method.**

        Parameters
        ----------
        adict : `dict`
            The keys for this dictionary have to be an instance of a downloader client.
            The values for each client are a dictionary that has keys of `~sunpy.net.attr.Attr`.
            Each key should have a list of tuples as it value.
            Each tuple should be a pair of strings.
            First string is the attribute name you want to register
            Second string is the description of the attribute.

        Examples
        --------
        The first import is to make this example work, it should not be used otherwise

        >>> from sunpy.net.dataretriever import GenericClient
        >>> from sunpy.net import attr, attrs
        >>> attr.Attr.update_values({GenericClient : {
        ...                          attrs.Instrument: [('AIA', 'AIA is in Space.'),
        ...                                             ('HMI', 'HMI is next to AIA.')]}})   # doctest: +SKIP
        >>> attr.Attr._attr_registry[attrs.Instrument]  # doctest: +SKIP
        attr(name=['aia', 'hmi'], client=['Generic', 'Generic'], name_long=['AIA', 'HMI'],
        desc=['AIA is in Space.', 'HMI is next to AIA.'])
        >>> attr.Attr._attr_registry[attrs.Instrument].name  # doctest: +SKIP
        ['aia', 'hmi']
        >>> attr.Attr._attr_registry[attrs.Instrument].name_long  # doctest: +SKIP
        ['AIA', 'HMI']
        >>> attr.Attr._attr_registry[attrs.Instrument].desc  # doctest: +SKIP
        ['AIA is in Space.', 'HMI is next to AIA.']
        >>> attrs.Instrument.aia, attrs.Instrument.hmi  # doctest: +SKIP
        (<sunpy.net.attrs.Instrument(AIA: AIA is in Space.) object at 0x...>,
        <sunpy.net.attrs.Instrument(HMI: HMI is next to AIA.) object at 0x...>)
        """
        for client, attr_dict in adict.items():
            for attr, attr_values in attr_dict.items():
                if not np.iterable(attr_values) or isinstance(attr_values, str):
                    raise ValueError(f"Invalid input value: {attr_values} for key: {repr(attr)}. "
                                     "The value is not iterable or just a string.")

                attr_tuple = cls._attr_registry[attr]

                for pair in attr_values:
                    if len(pair) > 2:
                        raise ValueError(f'Invalid length (!=2) for values: {attr_values}.')
                    elif len(pair) == 1:
                        if pair[0] != "*":
                            raise ValueError(
                                f'Invalid value given for * registration: {attr_values}.')
                        # Special case handling for * aka all values allowed.
                        pair = ["all", "All values of this type are supported."]

                    # Sanitize part one: Check if the name has a number in it
                    number_match = NUMBER_REGEX.match(pair[0])
                    p = inflect.engine()
                    try:
                        number_str = number_match.group(1)
                        name = p.number_to_words(number_str)
                        if number_str != number_match.string:
                            name = name + "_" + number_match.string[number_match.end(1):]
                    except AttributeError:
                        name = pair[0]

                    # Sanitize part two: remove punctuation and replace it with _
                    name = re.sub(f'[{re.escape(string.punctuation)}]', '_', name)
                    # Sanitize name, we remove all special characters
                    name = ''.join(char for char in name
                                   if char.isidentifier() or char.isnumeric())
                    # Make name lower case
                    name = name.lower()

                    if keyword.iskeyword(name):
                        # Attribute name has been appended with `_`
                        # to make it a valid identifier since its a python keyword.
                        name = name + '_'

                    attr_tuple[0].append(name)
                    attr_tuple[1].append(client.__name__.replace("Client", ""))
                    attr_tuple[2].append(pair[0])
                    attr_tuple[3].append(pair[1])


class DataAttr(Attr):
    """
    A base class for attributes classes which contain data.

    This is to differentiate them from classes like `sunpy.net.attr.AttrAnd` or
    the base `sunpy.net.attr.Attr` class which do not. The motivation for this
    distinction is to make it easier for walkers to match all classes which are
    not user specified Attrs.
    """
    def __new__(cls, *args, **kwargs):
        if cls is DataAttr:
            raise TypeError("You should not directly instantiate DataAttr, only it's subclasses.")
        return super().__new__(cls)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        # Because __new__() is defined, this will block natural introspection of the arguments for
        # __init__() in all subclasses because the signature of __new__() takes precedence over the
        # signature of __init__(). We add a __new__() to all subclasses that do not explicitly
        # define it with a signature that matches __init__().
        if '__new__' not in cls.__dict__:
            unsigned_new = cls.__new__  # the inherited __new__()

            def signed_new(cls, *args, **kwargs):
                return unsigned_new(cls, *args, **kwargs)
            signed_new.__signature__ = inspect.signature(cls.__init__)
            cls.__new__ = signed_new


class DummyAttr(Attr):
    """
    Empty attribute. Useful for building up queries. Returns other
    attribute when ORed or ANDed. It can be considered an empty query
    that you can use as an initial value if you want to build up your
    query in a loop.

    So, if we wanted an attr matching all the time intervals between the times
    stored as (from, to) tuples in a list, we could do.

    .. code-block:: python

       attr = DummyAttr()
       for from, to in times:
           attr |= Time(from, to)
    """

    def __and__(self, other):
        return other

    def __or__(self, other):
        return other

    def collides(self, other):
        return False

    def __hash__(self):
        return hash(None)

    def __eq__(self, other):
        return isinstance(other, DummyAttr)


class SimpleAttr(DataAttr):
    """
    An attribute that only has a single value.

    This type of attribute is not a composite and has a single value such as
    ``Instrument('EIT')``.

    Parameters
    ----------
    value : `object`
       The value for the attribute to hold.
    """

    def __init__(self, value):
        super().__init__()
        self.value = value

    def collides(self, other):
        return isinstance(other, self.__class__)

    def __repr__(self):
        obj_placeholder = " object "
        attr_reg = AttrMeta._attr_registry[self.__class__]
        new_repr = object.__repr__(self).split(obj_placeholder)
        # If somehow the idx isn't in the attr reg,
        # we still want it to print it repr without error.
        try:
            idx = attr_reg.name_long.index(self.value)
            obj_value_repr = f"({self.value}: {attr_reg.desc[idx]})"
        except ValueError:
            obj_value_repr = f": {self.value}"
        new_repr = [new_repr[0], obj_value_repr, obj_placeholder, new_repr[1]]
        return textwrap.fill("".join(new_repr), 100)

    @property
    def type_name(self):
        return self.__class__.__name__.lower()


class AttrComparison(DataAttr):
    """
    Allows a Attr to have a value and a comparison operator.
    """

    def __init__(self, name, operator, value):
        super().__init__()
        self.name = name
        self.operator = operator
        self.value = value

    def collides(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.operator == other.operator and self.name == other.name


class ComparisonParamAttrWrapper:
    def __init__(self, name):
        self.name = name

    def __lt__(self, other):
        return AttrComparison(self.name, '<', other)

    def __le__(self, other):
        return AttrComparison(self.name, '<=', other)

    def __gt__(self, other):
        return AttrComparison(self.name, '>', other)

    def __ge__(self, other):
        return AttrComparison(self.name, '>=', other)

    def __eq__(self, other):
        return AttrComparison(self.name, '=', other)

    def __ne__(self, other):
        return AttrComparison(self.name, '!=', other)

    def collides(self, other):
        return isinstance(other, ComparisonParamAttrWrapper)


class Range(DataAttr):
    """
    An attribute that represents a range of a value.

    This type of attribute would be applicable for types like Wavelength or Time.
    The range is inclusive of both the min and max.

    Parameters
    ----------
    min_ : `object`
        The lower bound of the range.
    max_ : `object`
        The upper bound of the range.
    """

    def __init__(self, min_, max_):
        self.min = min_
        self.max = max_

        super().__init__()

    def __xor__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented

        new = DummyAttr()
        if self.min < other.min:
            new |= type(self)(self.min, min(other.min, self.max))
        if other.max < self.max:
            new |= type(self)(other.max, self.max)
        return new

    def __contains__(self, other):
        if isinstance(other, Range):
            return self.min <= other.min and self.max >= other.max
        else:
            return self.min <= other <= self.max


class AttrAnd(Attr):
    """ Attribute representing attributes ANDed together. """

    def __init__(self, attrs):
        super().__init__()
        self.attrs = attrs

    def __and__(self, other):
        if any(other.collides(elem) for elem in self.attrs):
            return NotImplemented
        if isinstance(other, AttrAnd):
            return AttrAnd([*self.attrs, *other.attrs])
        if isinstance(other, AttrOr):
            return AttrOr([elem & self for elem in other.attrs])
        return AttrAnd([*self.attrs, other])

    __rand__ = __and__

    def __repr__(self):
        return f"<AttrAnd({self.attrs!r})>"

    def __eq__(self, other):
        if not isinstance(other, AttrAnd):
            return False
        return set(self.attrs) == set(other.attrs)

    def __hash__(self):
        return hash(frozenset(self.attrs))

    def collides(self, other):
        return any(elem.collides(other) for elem in self.attrs)


class AttrOr(Attr):
    """ Attribute representing attributes ORed together. """

    def __init__(self, attrs):
        super().__init__()
        self.attrs = attrs

    def __or__(self, other):
        if isinstance(other, AttrOr):
            return AttrOr([*self.attrs, *other.attrs])
        return AttrOr([*self.attrs, other])

    __ror__ = __or__

    def __and__(self, other):
        return AttrOr([elem & other for elem in self.attrs])

    __rand__ = __and__

    def __xor__(self, other):
        new = AttrOr([])
        for elem in self.attrs:
            try:
                new |= elem ^ other
            except TypeError:
                pass
        return new

    def __contains__(self, other):
        for elem in self.attrs:
            try:
                if other in elem:
                    return True
            except TypeError:
                pass
        return False

    def __repr__(self):
        return f"<AttrOr({self.attrs!r})>"

    def __eq__(self, other):
        if not isinstance(other, AttrOr):
            return False
        return set(self.attrs) == set(other.attrs)

    def __hash__(self):
        return hash(frozenset(self.attrs))

    def collides(self, other):
        return all(elem.collides(other) for elem in self.attrs)


# This appears to only be used as a base type for the Walker, i.e. a common
# denominator for the walker to convert to whatever the output of the walker is
# going to be.
class ValueAttr(DataAttr):
    def __init__(self, attrs):
        super().__init__()
        self.attrs = attrs

    def __repr__(self):
        return f"<ValueAttr({self.attrs!r})>"

    def __hash__(self):
        return hash(frozenset(self.attrs.items()))

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.attrs == other.attrs

    def collides(self, other):
        if not isinstance(other, self.__class__):
            return False
        return any(k in other.attrs for k in self.attrs)


class AttrWalker:
    """
    Traverse the Attr tree and convert it to a different representation.

    The ``AttrWalker`` can walk a complex tree of attrs and represent that tree
    in a way that is useful to the client using the attrs. For the VSO client
    it generates a ``VSOQueryResponseTable``.

    The walker has three core operations that can be applied to the tree, all
    of these are functions which are applied to one or more
    `~sunpy.net.attr.Attr` types, using conditional dispatch based on type.

    * creators: Creators when given an `~sunpy.net.attr.Attr` return a new object.
    * appliers: Appliers process an `~sunpy.net.attr.Attr` type and modify any
      arguments passed.
    * converters: Converters convert types unknown to any other creator or
      appliers to types known by them. They take in an `~sunpy.net.attr.Attr`
      type and return a different one.

    """

    def __repr__(self):
        creators = list(self.createmm.registry.keys())
        appliers = list(self.applymm.registry.keys())
        return f"""{super().__repr__()}
Registered creators:\n {creators}\n
Registered appliers:\n {appliers}"""

    @staticmethod
    def _unknown_type_apply(*args, **kwargs):
        raise TypeError(
            f"{args[1]} or any of its parents have not been registered using "
            "add_applier() with the AttrWalker")

    @staticmethod
    def _unknown_type_create(*args, **kwargs):
        raise TypeError(
            f"{args[1]} or any of its parents have not been registered using "
            "add_creator() with the AttrWalker")

    def __init__(self):
        self.applymm = seconddispatch(self._unknown_type_apply)
        self.createmm = seconddispatch(self._unknown_type_create)

    def create(self, *args, **kwargs):
        """
        Call the create function(s) matching the arguments to this method.
        """
        return self.createmm(self, *args, **kwargs)

    def apply(self, *args, **kwargs):
        """
        Call the apply function(s) matching the arguments to this method.
        """
        return self.applymm(self, *args, **kwargs)

    def add_creator(self, *types):
        """
        Register all specified types with this function for the ``.create`` method.
        """
        def _dec(fun):
            for type_ in types:
                self.createmm.register(type_, fun)
            return fun
        return _dec

    def add_applier(self, *types):
        """
        Register all specified types with this function for the ``.apply`` method.
        """
        def _dec(fun):
            for type_ in types:
                self.applymm.register(type_, fun)
            return fun
        return _dec

    def add_converter(self, *types):
        """
        Register a function to convert the specified type into a known type for create and apply.

        After a converter is run, create or apply will be called again with the new types.
        """
        def _dec(fun):
            for type_ in types:
                self.applymm.register(type_, self._cv_apply(fun))
                self.createmm.register(type_, self._cv_create(fun))
            return fun
        return _dec

    def _cv_apply(self, fun):
        """
        Call the converter and then re-call apply.
        """
        def _fun(*args, **kwargs):
            args = list(args)
            args[1] = fun(args[1])
            return self.applymm(*args, **kwargs)
        return _fun

    def _cv_create(self, fun):
        """
        Call the converter and then re-call convert.
        """
        def _fun(*args, **kwargs):
            args = list(args)
            args[1] = fun(args[1])
            return self.createmm(*args, **kwargs)
        return _fun


def and_(*args):
    """ Trick operator precedence.

    and_(foo < bar, bar < baz)
    """
    value = DummyAttr()
    for elem in args:
        value &= elem
    return value


def or_(*args):
    """ Trick operator precedence.

    or_(foo < bar, bar < baz)
    """
    value = DummyAttr()
    for elem in args:
        value |= elem
    return value



"""
=========================================
Masking HMI based on the intensity of AIA
=========================================

In this example we will demonstrate how to mask out regions within
a HMI image based on the intensity values of AIA.
"""
# sphinx_gallery_thumbnail_number = 3

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import Normalize
from skimage.measure import label, regionprops

import astropy.units as u
from astropy.coordinates import SkyCoord

import sunpy.map
from sunpy.data.sample import AIA_171_IMAGE, HMI_LOS_IMAGE

###############################################################################
# We will use an AIA 171 image from the sample data and crop it to capture a region of interest.

aia = sunpy.map.Map(AIA_171_IMAGE)
aia = aia.submap(
    bottom_left=SkyCoord(-250, 200, unit=u.arcsec, frame=aia.coordinate_frame),
    width=500 * u.arcsec,
    height=400 * u.arcsec,
)

###############################################################################
# Next we create the HMI map and crop it to the same field of view as the AIA image.

hmi = sunpy.map.Map(HMI_LOS_IMAGE)
hmi = hmi.submap(aia.bottom_left_coord, top_right=aia.top_right_coord)

###############################################################################
# We then call the :meth:`~sunpy.map.GenericMap.reproject_to` to reproject the AIA Map
# to have exactly the same grid as the HMI Map.
# We choose to reproject the AIA data to the HMI grid, rather than the reverse,
# to avoid interpolating the LOS HMI magnetic field data.
# This is because the range of the HMI data includes both positive and negative values and interpolation can destroy small scale variations in the LOS magnetic field which may be important in some scientific contexts.

aia = aia.reproject_to(hmi.wcs)
aia.nickname = 'AIA'

###############################################################################
# Now we will identify separate regions below a threshold in the AIA Map.
# In this case, we want the darker patches that have pixel values below 200.
# Then, using `skimage`, we can :func:`~skimage.measure.label`
# and calculate the properties of each region using :func:`~skimage.measure.regionprops`.

segmented = aia.data < 200
labeled = label(segmented)
regions = regionprops(labeled, hmi.data)
# We want the largest region, so we will sort by descending order in size.
regions = sorted(regions, key=lambda r: r.area, reverse=True)

###############################################################################
# Now to plot and label the first 7 regions seen in AIA with the region "0" being the largest.

fig = plt.figure()
ax = fig.add_subplot(projection=aia)
aia.plot(axes=ax)
aia.draw_contours(axes=ax, levels=200, colors="r")
for i in range(7):
    plt.text(*np.flip(regions[i].centroid), str(i), color="w", ha="center", va="center")

###############################################################################
# Now let's plot those same regions on the HMI Map.

fig = plt.figure()
ax = fig.add_subplot(projection=hmi)
im = hmi.plot(axes=ax, cmap="hmimag", norm=Normalize(-1500, 1500))
aia.draw_contours(axes=ax, levels=200, colors="r")
fig.colorbar(im)

###############################################################################
# Now we have the regions, we need to create a new HMI map that masks out everything but the largest region.
# To do so, we need to create the mask from the bounding box returned by `skimage`.
# Note that we can do this from the thresholded region only because our AIA and HMI images are on the same pixel grid after reprojecting the AIA image.

bbox = regions[0].bbox
mask = np.ones_like(hmi.data, dtype=bool)
mask[bbox[0]: bbox[2], bbox[1]: bbox[3]] = ~regions[0].image
hmi_masked = sunpy.map.Map((hmi.data, hmi.meta), mask=mask)

###############################################################################
# We can then plot the largest HMI region.

fig = plt.figure()
ax = fig.add_subplot(projection=hmi_masked)
im = hmi_masked.plot(axes=ax, cmap="hmimag", norm=Normalize(-1500, 1500))
fig.colorbar(im)

###############################################################################
# Finally, we can plot the distribution of HMI LOS magnetic field for only the unmasked values in the largest region shown above.

fig = plt.figure()
ax = fig.add_subplot()
ax.hist(hmi_masked.data[~hmi_masked.mask], bins='auto', histtype='step')
ax.set_ylabel('Number of Pixels')
ax.set_xlabel(f'LOS Magnetic Field [{hmi.unit:latex_inline}]')

plt.show()



"""
Test mapsequence functionality
"""

import numpy as np
import pytest

import astropy.units as u
from astropy.tests.helper import assert_quantity_allclose
from astropy.visualization import ImageNormalize

import sunpy
import sunpy.data.test
import sunpy.map
from sunpy.data.test import get_test_filepath
from sunpy.tests.helpers import figure_test, skip_glymur
from sunpy.util.metadata import MetaDict


@pytest.fixture
def mapsequence_all_the_same(aia171_test_map):
    """ Simple `sunpy.map.mapsequence` for testing."""
    return sunpy.map.Map([aia171_test_map, aia171_test_map], sequence=True)


@pytest.fixture
def mapsequence_different_maps(aia171_test_map, eit_test_map):
    """
    Simple `sunpy.map.mapsequence` for testing, in which there are different maps
    """
    return sunpy.map.Map([aia171_test_map, eit_test_map], sequence=True)


@pytest.fixture
def mapsequence_all_the_same_all_have_masks(aia171_test_map_with_mask):
    """ Simple `sunpy.map.mapsequence` for testing, in which all the maps have
    masks."""
    return sunpy.map.Map([aia171_test_map_with_mask, aia171_test_map_with_mask], sequence=True)


@pytest.fixture
def mapsequence_all_the_same_some_have_masks(aia171_test_map, aia171_test_map_with_mask):
    """ Simple `sunpy.map.mapsequence` for testing, in which at least some of the
    maps have masks."""
    return sunpy.map.Map([aia171_test_map_with_mask, aia171_test_map_with_mask, aia171_test_map], sequence=True)


@pytest.fixture()
def mapsequence_different(aia171_test_map):
    """ Mapsequence allows that the size of the image data in each map be
    different. This mapsequence contains such maps."""
    return sunpy.map.Map([aia171_test_map, aia171_test_map.superpixel((4, 4) * u.pix)], sequence=True)


def test_as_array_different_shapes(mapsequence_different):
    # Should raise a ValueError if the mapsequence has differently shaped maps
    with pytest.raises(ValueError, match='Not all maps have the same shape.'):
        mapsequence_different.data


def test_as_array_no_masks(mapsequence_all_the_same):
    # Test the case when none of the maps have a mask
    returned_array = mapsequence_all_the_same.data
    assert isinstance(returned_array, np.ndarray)
    assert returned_array.shape == (128, 128, 2)
    assert mapsequence_all_the_same.mask is None


def test_as_array_all_masks(mapsequence_all_the_same_all_have_masks):
    # Test the case when all the maps have masks
    data = mapsequence_all_the_same_all_have_masks.data
    assert data.shape == (128, 128, 2)
    mask = mapsequence_all_the_same_all_have_masks.mask
    assert mask.shape == (128, 128, 2)
    assert mask.dtype == bool


def test_as_array_some_masks(mapsequence_all_the_same_some_have_masks):
    data = mapsequence_all_the_same_some_have_masks.data
    assert data.shape == (128, 128, 3)
    mask = mapsequence_all_the_same_some_have_masks.mask
    assert mask.shape == (128, 128, 3)
    assert np.all(mask[0:2, 0:3, 0])
    assert np.all(mask[0:2, 0:3, 1])
    assert np.all(np.logical_not(mask[0:2, 0:3, 2]))


def test_all_meta(mapsequence_all_the_same):
    meta = mapsequence_all_the_same.meta
    assert len(meta) == 2
    assert np.all(np.asarray([isinstance(h, MetaDict) for h in meta]))
    assert np.all(np.asarray(
        [meta[i] == mapsequence_all_the_same[i].meta for i in range(0, len(meta))]
    ))


def test_repr(mapsequence_all_the_same, mapsequence_different_maps):
    """
    Tests that overridden __repr__ functionality works as expected. Test
    for mapsequence of same maps as well that of different maps.
    """
    # Test the case of MapSequence having same maps
    expected_out = 'MapSequence of 2 elements, with maps from AIAMap'
    obtained_out = repr(mapsequence_all_the_same)
    assert obtained_out.startswith(object.__repr__(mapsequence_all_the_same))
    assert len(mapsequence_all_the_same) == 2
    assert expected_out in obtained_out

    # Test the case of MapSequence having different maps
    expected_out1 = 'MapSequence of 2 elements, with maps from AIAMap, EITMap'
    expected_out2 = 'MapSequence of 2 elements, with maps from EITMap, AIAMap'
    obtained_out = repr(mapsequence_different_maps)
    assert obtained_out.startswith(object.__repr__(mapsequence_different_maps))
    assert len(mapsequence_different_maps) == 2
    assert expected_out1 in obtained_out or expected_out2 in obtained_out


def test_repr_html(mapsequence_all_the_same):
    html_string = mapsequence_all_the_same._repr_html_()
    for m in mapsequence_all_the_same.maps:
        assert m._repr_html_() in html_string


def test_quicklook(mocker, mapsequence_all_the_same):
    mockwbopen = mocker.patch('webbrowser.open_new_tab')
    mapsequence_all_the_same.quicklook()
    # Check that the mock web browser was opened with a file URL
    mockwbopen.assert_called_once()
    file_url = mockwbopen.call_args[0][0]
    assert file_url.startswith('file://')
    # Open the file specified in the URL and confirm that it contains the HTML
    with open(file_url[7:]) as f:
        html_string = f.read()
    for m in mapsequence_all_the_same.maps:
        assert m._repr_html_() in html_string


@figure_test
def test_norm_animator(hmi_test_map):
    seq = sunpy.map.Map([hmi_test_map, hmi_test_map], sequence=True)
    ani = seq.peek()
    ani._slider_changed(1, ani.sliders[ani.active_slider]._slider)
    return ani.fig


@figure_test
def test_map_sequence_plot(aia171_test_map, hmi_test_map):
    seq = sunpy.map.Map([aia171_test_map, hmi_test_map], sequence=True)
    seq.plot()


@figure_test
def test_map_sequence_plot_custom_cmap_norm(aia171_test_map, hmi_test_map):
    seq = sunpy.map.Map([aia171_test_map, hmi_test_map], sequence=True)
    animation = seq.plot(cmap='Greys', norm=ImageNormalize(vmin=0, vmax=100))
    animation._step()


def test_save(aia171_test_map, hmi_test_map, tmp_path):
    """
    Tests the MapSequence save function
    """
    seq = sunpy.map.Map([aia171_test_map, hmi_test_map], sequence=True)

    with pytest.raises(ValueError, match="'{index}' must be appear in the string"):
        seq.save("index", filetype='fits', overwrite=True)

    base_str = (tmp_path / "map").as_posix()
    seq.save(f"{base_str}_{{index:03}}.fits", filetype='auto', overwrite=True)

    test_seq = sunpy.map.Map(f"{base_str}_000.fits",
                             f"{base_str}_001.fits", sequence=True)

    assert isinstance(test_seq.maps[0], sunpy.map.sources.sdo.AIAMap)
    assert isinstance(test_seq.maps[1], sunpy.map.sources.sdo.HMIMap)

    assert test_seq.maps[0].meta.keys() == seq.maps[0].meta.keys()
    for k in seq.maps[0].meta:
        assert test_seq.maps[0].meta[k] == seq.maps[0].meta[k]
    assert_quantity_allclose(test_seq.maps[0].data, seq.maps[0].data)

    assert test_seq.maps[1].meta.keys() == seq.maps[1].meta.keys()
    for k in seq.maps[1].meta:
        assert test_seq.maps[1].meta[k] == seq.maps[1].meta[k]
    assert_quantity_allclose(test_seq.maps[1].data, seq.maps[1].data)


@figure_test
def test_map_sequence_plot_clip_interval(aia171_test_map):
    seq = sunpy.map.Map([aia171_test_map, aia171_test_map, aia171_test_map], sequence=True)
    # We want to blow the image out to make sure it is clipped on the test data
    animation = seq.plot(clip_interval=(5,75)*u.percent)
    animation._step()


@figure_test
def test_map_sequence_plot_custom_title(aia171_test_map):
    # Test that custom plot title shows up even when annotate=True
    seq = sunpy.map.Map([aia171_test_map, aia171_test_map], sequence=True)
    seq.plot(title='Custom Title', annotate=True)


@skip_glymur
def test_mapsequence_plot_uint8_norm():
    # We want to check the case for images with no norms set
    # This code used to fail in this case.
    coconuts = sunpy.map.Map([get_test_filepath("2013_06_24__17_31_30_84__SDO_AIA_AIA_193.jp2")]*10,sequence=True)
    [coconut.plot_settings.pop("norm") for coconut in coconuts]
    moving_coconut = coconuts.plot()
    moving_coconut._step()


@skip_glymur
def test_mapsequence_plot_uint8_norm_clip_interval():
    # This code used to fail in this case.
    coconuts = sunpy.map.Map([get_test_filepath("2013_06_24__17_31_30_84__SDO_AIA_AIA_193.jp2")]*10,sequence=True)
    [coconut.plot_settings.pop("norm") for coconut in coconuts]
    moving_coconut = coconuts.plot(clip_interval=(1, 99.99)*u.percent)
    moving_coconut._step()


@skip_glymur
def test_mapsequence_plot_set_norm_pass_vmin_vmax(aia171_test_map):
    # CHecking that passing in interval values works if the map has a norm
    coconuts = sunpy.map.Map([aia171_test_map]*10,sequence=True)
    moving_coconut = coconuts.plot(clip_interval=(1, 99.99)*u.percent)
    moving_coconut._step()
    moving_coconut = coconuts.plot(vmin=100, vmax=1000)
    moving_coconut._step()



#!/bin/env python

import sunpy
from sunpy.net.cdaweb.helpers import _update_cdaweb_dataset_data
from sunpy.net.jsoc import JSOCClient
from sunpy.net.vso import VSOClient
from sunpy.net.solarnet import SOLARNETClient

print(f"Updating the attrs json files using sunpy {sunpy.__version__}...")

print("Updating VSO json...")
VSOClient._update_vso_data()

print("Updating JSOC json...\nThis may take some time...")
JSOCClient._update_jsoc_data()

print("Updating SOLARNET json...\nThis may take some time...")
SOLARNETClient._update_solarnet_data()

print("Updating CDAWeb json...\nThis may take some time...")
_update_cdaweb_dataset_data()

print("Done. Don't forget to update the doctests.")



"""
This module provides a wrapper around the VSO API.
"""

import os
import copy
import json
import inspect
import datetime
import itertools
from pathlib import Path
from functools import partial
from urllib.error import URLError, HTTPError
from urllib.parse import urlencode
from urllib.request import Request, urlopen

import zeep

from sunpy import config, log
from sunpy.net import _attrs as core_attrs
from sunpy.net.attr import and_
from sunpy.net.base_client import BaseClient, QueryResponseRow
from sunpy.net.vso import attrs
from sunpy.net.vso.attrs import _walker as walker
from sunpy.util.exceptions import warn_connection, warn_user
from sunpy.util.net import parse_header, slugify
from sunpy.util.parfive_helpers import Downloader, Results
from .exceptions import (
    DownloadFailed,
    MissingInformation,
    MultipleChoices,
    NoData,
    UnknownMethod,
    UnknownStatus,
    UnknownVersion,
)
from .table_response import VSOQueryResponseTable
from .zeep_plugins import SunPyLoggingZeepPlugin

DEFAULT_URL_PORT = [
    {'url': 'http://docs.virtualsolar.org/WSDL/VSOi_rpc_literal.wsdl', 'port': 'nsoVSOi'},
    {'url': 'https://sdac.virtualsolar.org/API/VSOi_rpc_literal.wsdl', 'port': 'sdacVSOi'},
    # This connects to the same cgi as the first WSDL file, but if that file
    # isn't accessible and the SDAC cgi is unreachable, this might still work.
    {'url': 'https://sdac.virtualsolar.org/API/VSOi_rpc_literal.wsdl', 'port': 'nsoVSOi'},
]


class _Str(str):
    """
    Subclass of string that contains a meta attribute for the
    record_item associated with the file.
    """
    meta = None


def check_connection(url):
    try:
        return urlopen(url, timeout=15).getcode() == 200
    except (OSError, HTTPError, URLError) as e:
        warn_connection(f"Connection to {url} failed with error {e}. Retrying with different url and port.")
        return False


def check_cgi_connection(url):
    """
    At the moment there is no way to make an "are you alive" request to the
    cgi, so we just hit it with a HTTP get and it gives us back a 411 response.
    This is weird enough that it probably satisfies us for this check.
    """
    try:
        return urlopen(url, timeout=15).getcode() == 411
    except HTTPError as e:
        if e.code == 411:
            return True
        warn_connection(f"Connection to {url} failed with error {e}. Retrying with different url and port.")
        return False
    except (OSError, URLError) as e:
        warn_connection(f"Connection to {url} failed with error {e}. Retrying with different url and port.")
        return False


def get_online_vso_url():
    """
    Return the first VSO url and port combination that is online.
    """
    for mirror in DEFAULT_URL_PORT:
        if check_connection(mirror['url']):
            # Now we get the port URL from the WSDL and test that
            wsdl = zeep.wsdl.Document(mirror["url"], zeep.Transport())
            # I think that accessing "VSOiService" here is equivalent to the
            # set_ns_prefix call in the build_client function below
            url = wsdl.services["VSOiService"].ports[mirror["port"]].binding_options["address"]
            if not check_cgi_connection(url):
                continue
            return mirror


def build_client(url=None, port_name=None, **kwargs):
    """
    Construct a `zeep.Client` object to connect to VSO.

    Parameters
    ----------
    url : `str`
        The URL to connect to.
    port_name : `str`
        The "port" to use.
    **kwargs : `dict`
        All extra keyword arguments are passed to `zeep.Client`.

    Returns
    -------
    `zeep.Client`
    """
    if url is None and port_name is None:
        mirror = get_online_vso_url()
        if mirror is None:
            raise ConnectionError("No online VSO mirrors could be found.")
        url = mirror['url']
        port_name = mirror['port']
    elif url and port_name:
        if not check_connection(url):
            raise ConnectionError(f"Can't connect to url {url}")
    else:
        raise ValueError("Both url and port_name must be specified if either is.")

    if "plugins" not in kwargs:
        kwargs["plugins"] = [SunPyLoggingZeepPlugin()]

    client = zeep.Client(url, port_name=port_name, **kwargs)
    client.set_ns_prefix('VSO', 'http://virtualsolar.org/VSO/VSOi')
    return client


class VSOClient(BaseClient):
    """
    Provides access to query and download from Virtual Solar Observatory (VSO).

    Parameters
    ----------
    url : `str`, optional
        The VSO url to use. If not specified will use the first online known URL.

    port : `str`, optional
        The VSO port name to use. If not specified will use the first online known URL.

    api : `zeep.Client`, optional
        The `zeep.Client` instance to use for interacting with the VSO. If not
        specified one will be created.
    """
    method_order = [
        'URL-FILE_Rice', 'URL-FILE', 'URL-packaged', 'URL-TAR_GZ', 'URL-ZIP', 'URL-TAR',
    ]

    def __init__(self, url=None, port=None, api=None):
        if not isinstance(api, zeep.Client):
            api = build_client(url, port)
            if api is None:
                raise ConnectionError("Cannot find an online VSO mirror.")
        self.api = api

    def __deepcopy__(self, memo):
        """
        Copy the client but don't copy the API object.
        """
        memo[id(self.api)] = self.api
        deepcopy_method = self.__deepcopy__
        self.__deepcopy__ = None
        cp = copy.deepcopy(self, memo)
        self.__deepcopy__ = deepcopy_method
        cp.__deepcopy__ = deepcopy_method
        return cp

    def make(self, atype, **kwargs):
        """
        Create a new SOAP object.
        """
        obj = self.api.get_type(f"VSO:{atype}")
        return obj(**kwargs)


    def search(self, *query):
        """
        Query data from the VSO with the new API. Takes a variable number
        of attributes as parameter, which are chained together using AND.

        Returns
        -------
        out : `~sunpy.net.vso.table_response.VSOQueryResponseTable`
            Matched items. Return value is of same type as the one of
            :meth:`VSOClient.search`.
        """
        query = and_(*query)
        QueryRequest = self.api.get_type('VSO:QueryRequest')
        VSOQueryResponse = self.api.get_type('VSO:QueryResponse')
        responses = []
        exceptions = []
        for block in walker.create(query, self.api):
            try:
                query_response = self.api.service.Query(
                    QueryRequest(block=block)
                )
                for resp in query_response:
                    if resp["error"]:
                        warn_user(resp["error"])
                responses.append(
                    VSOQueryResponse(query_response)
                )
            except Exception as ex:
                exceptions.append(ex)

        responses = self.merge(responses)
        response = VSOQueryResponseTable.from_zeep_response(responses, client=self)

        for ex in exceptions:
            response.add_error(ex)

        return response

    def merge(self, queryresponses):
        """ Merge responses into one. """
        if len(queryresponses) == 1:
            return queryresponses[0]

        fileids = set()
        providers = {}

        for queryresponse in queryresponses:
            for provideritem in queryresponse.provideritem:
                provider = provideritem.provider
                if not hasattr(provideritem, 'record'):
                    continue
                if not hasattr(provideritem.record, 'recorditem'):
                    continue
                if provideritem.provider not in providers:
                    providers[provider] = provideritem
                    fileids |= {
                        record_item.fileid
                        for record_item in provideritem.record.recorditem
                    }
                else:
                    for record_item in provideritem.record.recorditem:
                        if record_item.fileid not in fileids:
                            fileids.add(record_item.fileid)
                            providers[provider].record.recorditem.append(
                                record_item
                            )
                            providers[provider].no_of_records_found += 1
                            providers[provider].no_of_records_returned += 1
        return self.make('QueryResponse',
                         provideritem=list(providers.values()))

    @staticmethod
    def mk_filename(pattern, queryresponserow, resp, url):
        """
        Generate the best possible (or least-worse) filename for a VSO download.

        * Use the ``content-disposition`` header.
        * Use ``fileid`` to generate a file name if content-disposition fails
        * If everything else fails use the last segment of the URL and hope.
        """
        name = None
        if resp:
            cdheader = resp.headers.get("Content-Disposition", None)
            if cdheader:
                _, params = parse_header(cdheader)
                name = params.get('filename', "")
                # Work around https://github.com/sunpy/sunpy/issues/3372
                if name.count('"') >= 2:
                    name = name.split('"')[1]

        # This is a hack to to prevent IRIS data from being labelled as XML files
        if name is None and "VOEvent_IRIS" not in queryresponserow['fileid']:
            # Advice from the VSO is to fallback to providerid + fileid for a filename
            # As it's possible multiple providers give the same fileid.
            # However, I haven't implemented this yet as it would be a breaking
            # change to the filenames we expect.
            fileid = queryresponserow['fileid']

            # Some providers make fileid a path
            # Some also don't specify a file extension, but not a lot we can do
            # about that.
            name = fileid.split("/")[-1]

        # If somehow we have got this far with an empty string, fallback to url segment
        if not name:
            name = url.split('/')[-1]

        # Remove any not-filename appropriate characters
        name = slugify(name)

        # If absolutely everything else fails make a filename based on download time
        if not name:
            name = f"vso_file_{datetime.datetime.now().strftime('%Y%m%d%H%M%S%f')}"

        fname = pattern.format(file=name,
                               **queryresponserow.response_block_map)

        return fname

    def fetch(self, query_response, path=None, methods=None, site=None,
              progress=True, overwrite=False, downloader=None, wait=True, **kwargs):
        """
        Download data specified in the query_response.

        Parameters
        ----------
        query_response : sunpy.net.vso.VSOQueryResponseTable
            VSOQueryResponseTable containing the items to be downloaded.
        path : str
            Specify where the data is to be downloaded. Can refer to arbitrary
            fields of the QueryResponseItem (instrument, source, time, ...) via
            string formatting, moreover the file-name of the file downloaded can
            be referred to as file, e.g.
            "{source}/{instrument}/{time.start}/{file}".
        methods : `list` of `str`
            Download methods, defaults to URL-FILE_Rice then URL-FILE.
            Methods are a concatenation of one PREFIX followed by any number of
            SUFFIXES i.e. ``PREFIX-SUFFIX_SUFFIX2_SUFFIX3``.
            The full list of
            `PREFIXES <https://sdac.virtualsolar.org/cgi/show_details?keyword=METHOD_PREFIX>`__
            and `SUFFIXES <https://sdac.virtualsolar.org/cgi/show_details?keyword=METHOD_SUFFIX>`__
            are listed on the VSO site.
        site : str
            There are a number of caching mirrors for SDO and other
            instruments, some available ones are listed below.

            =============== ========================================================
            NSO             National Solar Observatory, Tucson (US)
            SAO  (aka CFA)  Smithonian Astronomical Observatory, Harvard U. (US)
            SDAC (aka GSFC) Solar Data Analysis Center, NASA/GSFC (US)
            ROB             Royal Observatory of Belgium (Belgium)
            MPS             Max Planck Institute for Solar System Research (Germany)
            UCLan           University of Central Lancashire (UK)
            IAS             Institut Aeronautique et Spatial (France)
            KIS             Kiepenheuer-Institut fur Sonnenphysik Germany)
            NMSU            New Mexico State University (US)
            =============== ========================================================

        progress : `bool`, optional
            If `True` show a progress bar showing how many of the total files
            have been downloaded. If `False`, no progress bars will be shown at all.
        overwrite : `bool` or `str`, optional
            Determine how to handle downloading if a file already exists with the
            same name. If `False` the file download will be skipped and the path
            returned to the existing file, if `True` the file will be downloaded
            and the existing file will be overwritten, if ``'unique'`` the filename
            will be modified to be unique.
        downloader : `parfive.Downloader`, optional
            The download manager to use.
        wait : `bool`, optional
            If `False` ``downloader.download()`` will not be called. Only has
            any effect if ``downloader`` is not `None`.

        Returns
        -------
        out : `parfive.Results`
            Object that supplies a list of filenames and any errors.

        Examples
        --------
        >>> files = fetch(qr) # doctest:+SKIP
        """
        if path is None:
            path = Path(config.get('downloads', 'download_dir')) / '{file}'
        elif isinstance(path, str | os.PathLike) and '{file}' not in str(path):
            path = Path(path) / '{file}'
        else:
            path = Path(path)
        path = path.expanduser()

        dl_set = True
        if not downloader:
            dl_set = False
            downloader = Downloader(progress=progress, overwrite=overwrite)

        if isinstance(query_response, list):
            query_response = VSOQueryResponseTable.from_zeep_response(
                query_response,
                client=self,
        )
        if isinstance(query_response, QueryResponseRow):
            query_response = query_response.as_table()

        if not len(query_response):
            return downloader.download() if wait else Results()

        # Adding the site parameter to the info
        info = {}
        if site is not None:
            info['site'] = site

        VSOGetDataResponse = self.api.get_type("VSO:VSOGetDataResponse")

        data_request = self.make_getdatarequest(query_response, methods, info)
        data_response = VSOGetDataResponse(self.api.service.GetData(data_request))

        err_results = self.download_all(data_response,
                                        methods,
                                        downloader,
                                        str(path),
                                        self.by_fileid(query_response),
                                        **kwargs)

        if dl_set and not wait:
            return err_results

        results = downloader.download()
        results += err_results
        results._errors += err_results.errors
        return results

    def make_getdatarequest(self, response, methods=None, info=None):
        """
        Make datarequest with methods from response.
        """
        # Pass back the Apache session ID to the VSO if it exists in the response
        for item in response:
            info_required = item.get("Info Required", None)
            if info_required is not None:
                info['required'] = item['Info Required']

        if methods is None:
            methods = self.method_order + ['URL']

        return self.create_getdatarequest(
            {g[0]['Provider']: list(g['fileid']) for g in response.group_by('Provider').groups},
            methods, info
        )

    def create_getdatarequest(self, maps, methods, info=None):
        """ Create datarequest from maps mapping data provider to
        fileids and methods,"""
        if info is None:
            info = {}

        if 'email' not in info:
            info['email'] = 'sunpy'

        # For the JSOC provider we need to make a DataRequestItem for each
        # series, not just one for the whole provider.

        # Remove JSOC provider items from the map
        jsoc = maps.pop('JSOC', [])
        # Make DRIs for everything that's not JSOC one per provider
        dris = [self.make('DataRequestItem', provider=k, fileiditem={'fileid': v})
                for k, v in maps.items()]

        def series_func(x):
            """ Extract the series from the fileid. """
            return x.split(':')[0]

        # Sort the JSOC fileids by series
        # This is a precursor to groupby as recommended by the groupby docs
        series_sorted = sorted(jsoc, key=series_func)
        # Iterate over the series and make a DRI for each.
        # groupby creates an iterator based on a key function, in this case
        # based on the series (the part before the first ':')
        for _, fileids in itertools.groupby(series_sorted, key=series_func):
            dris.append(self.make('DataRequestItem',
                                  provider='JSOC',
                                  fileiditem={'fileid': list(fileids)}))

        request = {'method': {'methodtype': methods},
                   'info': info,
                   'datacontainer': {'datarequestitem': dris}
                   }

        return self.make('VSOGetDataRequest', request=request)

    def download_all(self, response, methods, downloader, path, qr, info=None, **kwargs):
        results = Results()
        GET_VERSION = [
            ('0.8', (5, 8)),
            ('0.7', (1, 4)),
            ('0.6', (0, 3)),
        ]

        for dresponse in response.getdataresponseitem:
            for version, (from_, to) in GET_VERSION:
                if getattr(dresponse, version, '0.6') >= version:
                    break
            else:
                results.add_error('', UnknownVersion(dresponse))
                continue

            # If from_ and to are uninitialized, the else block of the loop
            # continues the outer loop and thus this code is never reached.
            code = (
                dresponse.status[from_:to]
                if getattr(dresponse, 'status', None) else '200'
            )
            if code == '200':
                for dataitem in dresponse.getdataitem.dataitem:

                    try:
                        self.download(
                            dresponse.method.methodtype[0],
                            dataitem.url,
                            downloader,
                            path,
                            qr[dataitem.fileiditem.fileid[0]],
                            **kwargs
                        )
                    except NoData:
                        results.add_error('', '', DownloadFailed(dresponse))
                        continue

            elif code == '300' or code == '412' or code == '405':
                if code == '300':
                    try:
                        methods = self.multiple_choices(
                            dresponse.method.methodtype, dresponse
                        )
                    except NoData:
                        results.add_error('', '', MultipleChoices(dresponse))
                        continue
                elif code == '412':
                    try:
                        info = self.missing_information(
                            info, dresponse.info
                        )
                    except NoData:
                        results.add_error('', '', MissingInformation(dresponse))
                        continue
                elif code == '405':
                    try:
                        methods = self.unknown_method(dresponse)
                    except NoData:
                        results.add_error('', '', UnknownMethod(dresponse))
                        continue

                files = []
                for dataitem in dresponse.getdataitem.dataitem:
                    files.extend(dataitem.fileiditem.fileid)

                request = self.create_getdatarequest(
                    {dresponse.provider: files}, methods, info
                )

                self.download_all(
                    self.api.service.GetData(request), methods, downloader, path,
                    qr, info, **kwargs
                )
            else:
                results.add_error('', '', UnknownStatus(dresponse))

        return results

    def download(self, method, url, downloader, *args, **kwargs):
        """ Enqueue a file to be downloaded, extra args are passed to ``mk_filename``"""
        if method.startswith('URL'):
            return downloader.enqueue_file(url, filename=partial(self.mk_filename, *args))

        raise NoData

    @staticmethod
    def by_fileid(response):
        """
        Returns a dictionary of fileids
        corresponding to records in the response.
        """
        return {
            record['fileid']: record for record in response
        }

    def multiple_choices(self, choices, response):
        """ Override to pick between multiple download choices. """
        for elem in self.method_order:
            if elem in choices:
                return [elem]
        raise NoData

    def missing_information(self, info, field):
        """ Override to provide missing information. """
        raise NoData

    def unknown_method(self, response):
        """ Override to pick a new method if the current one is unknown. """
        raise NoData

    @classmethod
    def _can_handle_query(cls, *query):
        required = {core_attrs.Time}
        # Get all classes in core_attrs and attrs
        optional = {value for (name, value) in inspect.getmembers(core_attrs) if
                    name in core_attrs.__all__}
        optional.update(value for (name, value) in inspect.getmembers(attrs) if
                        name in attrs.__all__)
        return cls.check_attr_types_in_query(query, required, optional)

    @classmethod
    def _attrs_module(cls):
        return 'vso', 'sunpy.net.vso.attrs'

    def __del__(self):
        """
        Attempt to close the connection, but if it fails, continue.
        """
        try:
            self.api.transport.session.close()
        except Exception as e:
            log.debug(f"Failed to close VSO API connection with: {e}")

    @classmethod
    def register_values(cls):
        # We always use the local file for now.
        return cls.load_vso_values()

    @staticmethod
    def load_vso_values():
        """
        We take this list and register all the keywords as corresponding Attrs.

        Returns
        -------
        dict
            The constructed Attrs dictionary ready to be passed into Attr registry.
        """
        from sunpy.net import attrs as a

        here = os.path.dirname(os.path.realpath(__file__))
        with open(os.path.join(here, 'data', 'attrs.json')) as attrs_file:
            keyword_info = json.load(attrs_file)

        # Now to traverse the saved dict and give them attr keys.
        attrs = {}
        for key, value in keyword_info.items():
            attr = getattr(a, key.capitalize(), None)
            if attr is None:
                attr = getattr(a.vso, key.capitalize())
            attrs[attr] = value
        return attrs

    @staticmethod
    def _update_vso_data():
        """
        Makes a network call to the VSO API that returns what keywords they support.
        We take this list and register all the keywords as corresponding Attrs.
        """
        here = os.path.dirname(os.path.realpath(__file__))

        # Keywords we are after
        keywords = ["+detector", "+instrument", "+source", "+provider", "+physobs", "+level"]
        # Construct and format the request
        keyword_info = {}
        url = "https://vso1.nascom.nasa.gov/cgi-bin/registry_json.cgi"
        for keyword in keywords:
            data = urlencode({'fields': f"['{keyword}']".replace("'", '"')})
            req = Request(url=url + f"?{data}")
            response = urlopen(req)
            keyword_info[keyword.replace("+", "")] = json.loads(response.read())

        # Now to traverse the return and create attrs out of them.
        attrs = {}
        for key, value in keyword_info.items():
            attrs[key] = []
            for item in value:
                if item:
                    if key == "level":
                        attrs[key].append((str(item[key]), str(item[key])))
                    else:
                        attrs[key].append((str(item[key]), str(item[key+"_long"])))

        # Sort values within each attribute
        for attr in attrs:
            attrs[attr] = sorted(attrs[attr], key=lambda _list: _list[0])

        with open(os.path.join(here, 'data', 'attrs.json'), 'w') as attrs_file:
            json.dump(dict(sorted(attrs.items())), attrs_file, indent=2)

    @property
    def info_url(self):
        return 'https://sdac.virtualsolar.org/cgi/search\nData retrieval status: http://docs.virtualsolar.org/wiki/VSOHealthReport'



import pytest

import astropy.units as u

from sunpy.data.test import get_dummy_map_from_header, get_test_filepath
from sunpy.map.sources.sdo import HMISynopticMap
from sunpy.util.exceptions import SunpyMetadataWarning
from .helpers import _test_private_date_setters


@pytest.fixture
def hmi_synoptic():
    return get_dummy_map_from_header(get_test_filepath('hmi_synoptic.header'))


def test_fitstoHMISynoptic(hmi_synoptic):
    """Tests the creation of HMISynopticMap using FITS."""
    assert isinstance(hmi_synoptic, HMISynopticMap)


def test_is_datasource_for(hmi_synoptic):
    """Test the is_datasource_for method of HMISynopticMap.
    Note that header data to be provided as an argument
    can be a MetaDict object, which in this case is
    hmi.meta."""
    assert hmi_synoptic.is_datasource_for(hmi_synoptic.data, hmi_synoptic.meta)


def test_observatory(hmi_synoptic):
    """Tests the observatory property of the HMISynopticMap object."""
    assert hmi_synoptic.observatory == "SDO"


def test_measurement(hmi_synoptic):
    """Tests the measurement property of the HMISynopticMap object."""
    assert hmi_synoptic.measurement == "carrington"


def test_reference_date(hmi_synoptic):
    assert hmi_synoptic.reference_date.isot == "2018-11-09T12:30:52.000"


def test_date(hmi_synoptic):
    assert hmi_synoptic.date.isot == "2018-11-09T12:30:52.000"


def test_private_date_setters(hmi_synoptic):
    _test_private_date_setters(hmi_synoptic)


def test_unit(hmi_synoptic):
    assert hmi_synoptic.unit == u.G
    assert hmi_synoptic.unit == u.Unit("Mx/cm^2")
    assert hmi_synoptic.unit.to_string() == 'Mx / cm2'
    hmi_synoptic.meta['bunit'] = 'm'
    assert hmi_synoptic.unit == u.m


def test_wcs(hmi_synoptic):
    # Smoke test that WCS is valid and can transform from pixels to world coordinates
    with pytest.warns(SunpyMetadataWarning, match='Missing metadata for observer'):
        hmi_synoptic.pixel_to_world(0*u.pix, 0*u.pix)



"""
ASO-S Map subclass definitions.
"""

import astropy.units as u

from sunpy.coordinates import get_earth
from sunpy.map.mapbase import GenericMap, SpatialPair
from sunpy.time import parse_time

__all__ = ['HXIMap']


class HXIMap(GenericMap):
    """
    HXI Image Map Source.

    The Hard X-ray Imager (HXI, :cite:t:`Zhang_asos_mission_hxi_2019, Su_asos_hxi_simu_soft_2019`)
    is one of the three payloads of the Advanced Space-based Solar Observatory
    (ASO-S, :cite:t:`Gan_asos_mission_overview_2019, Gan_asos_issue_overview_2023`), which is designed
    to observe Hard X-Ray (HXR) spectra and images of solar flares. Having 91 subcollimators to
    modulate incident x-rays, HXI can obtain 91 modulation data and 45 visibilities to reconstruct
    images with a spatial resolution as high as ~3.1 arcsec.

    In addition, the energy ranges are ~15-300 keV for imaging and ~10-300 keV for spectra, more information
    please refer to :cite:t:`Su_asos_hxi_tests_calibrations_2024`.

    The HXI level 1 detector data, which is provided by the
    `ASO-S Data Center <http://aso-s.pmo.ac.cn/sodc/dataArchive.jsp>`__, can be further processed by
    the HXI `data analysis software <http://aso-s.pmo.ac.cn/sodc/analysisSoftware.jsp>`__
    to make hard X-ray images and export them as a FITS file.
    There may be more than one image included in the FITS file, from different time and energy intervals, which can be
    loaded by `~sunpy.map.Map` as a single `~sunpy.map.sources.HXIMap` (for one image) or a `list` containing
    multiple `~sunpy.map.sources.HXIMap`.

    It is suggested to pay attention to the updates in the data and software versions, and feel free to contact the
    ASO-S/HXI team if there are any difficulties in data processing.
    Some useful links are attached in the references section below.

    References
    ----------
    * `ASO-S Mission Page <http://aso-s.pmo.ac.cn/en_index.jsp>`__
    * `ASO-S Instruments <http://aso-s.pmo.ac.cn/english/sat/payload.jsp>`__
    * `ASO-S Data Center <http://aso-s.pmo.ac.cn/sodc/dataArchive.jsp>`__
    * `Data Analysis Guide <http://aso-s.pmo.ac.cn/sodc/analysisGuide.jsp>`__
    * `Data Analysis Software <http://aso-s.pmo.ac.cn/sodc/analysisSoftware.jsp>`__
    * `ASO-S/HXI Flare List <http://aso-s.pmo.ac.cn/hxi_flare/hxi_flare_list.html>`__
    * The tests and calibrations paper: :cite:t:`Su_asos_hxi_tests_calibrations_2024`
    """

    def __init__(self, data, header, **kwargs):
        super().__init__(data, header, **kwargs)
        self._nickname = self.meta.get('nickname', 'ASO-S/HXI')
        self.plot_settings['cmap'] = 'rhessi'

    @property
    def _default_observer_coordinate(self):
        if not (set(['HGLN_OBS','HGLT_OBS','DSUN_OBS']) < set(self.fits_header)):
            return  get_earth(self.reference_date)

    def _get_cmap_name(self):
        return "rhessi"

    @property
    def spatial_units(self):
        units = [self.meta.get('cunit1', None), self.meta.get('cunit2', None)]
        if self.meta['cunit1'] == 'arcsecs':
            units[0] = 'arcsec'
        if self.meta['cunit2'] == 'arcsecs':
            units[1] = 'arcsec'
        units = [None if unit is None else u.Unit(unit.lower()) for unit in units]
        return SpatialPair(units[0], units[1])

    @property
    def coordinate_system(self):
        ctypes = [self.meta['ctype1'], self.meta['ctype2']]
        if ctypes[0] == 'solar_x':
            ctypes[0] = 'HPLN-TAN'
        if ctypes[1] == 'solar_y':
            ctypes[1] = 'HPLT-TAN'
        return SpatialPair(*ctypes)

    @property
    def date_start(self):
        """
        DATE-OBS is the start_date.
        Here do some thing to deal with date string like '01-May-23 13:08:16.962' or
        ' 1-May-2023 13:08:16.962' (the latter one is output from SSWIDL's map2fits)
        """
        date_obs = self.meta.get('DATE-OBS').strip()
        try:
            return parse_time(date_obs)
        except ValueError:
            tmp = date_obs.split(" ")[0]
            if len(tmp) == 9 and len(tmp.split("-")[-1]) == 2:
                date_obs = date_obs[:7] + '20'+date_obs[7:]
            return parse_time(date_obs)

    @property
    def date_end(self):
        return self.date_start + self.meta.get('exptime')*u.second

    @property
    def _date_obs(self):
        return self.date_start

    @property
    def reference_date(self):
        return self.date_start

    @property
    def unit(self):
        unit = self.meta.get("unit")
        if unit is None:
            # Fall back to BUNIT
            return super().unit

        unit = unit.replace("photons", "photon")
        unit = unit.replace("asec", "arcsec")
        return self._parse_fits_unit(unit)

    @property
    def waveunit(self):
        return u.Unit(self.meta.get("waveunit", 'keV'))

    @property
    def wavelength(self):
        waves = [self.meta.get('energy_l', None), self.meta.get('energy_h', None)]
        if None in waves:
            return None
        else:
            return waves*self.waveunit

    @property
    def observatory(self):
        return self.meta.get('observatory', 'ASO-S')

    @property
    def instrument(self):
        return self.meta.get('instrume', 'HXI')

    @property
    def detector(self):
        return self.meta.get('detector', 'HXI')

    @classmethod
    def is_datasource_for(cls, data, header, **kwargs):
        """
        Determines if header corresponds to a HXI image.
        """
        return 'HXI' in header.get('ORIGIN', '') or header.get('INSTRUME') == 'HXI'



"""
=====================================================
Creating a time-distance plot from a sequence of maps
=====================================================

This example showcases how you can use :func:`sunpy.map.pixelate_coord_path`
and :func:`sunpy.map.sample_at_coords` on a sequence of images to create a
time-distance diagram accounting for solar differential rotation using
`sunpy.coordinates.propagate_with_solar_surface` and dealing with off-disk
pixels using `sunpy.coordinates.screens.SphericalScreen`
"""
# sphinx_gallery_thumbnail_number = -1
import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import numpy as np

import astropy.units as u
from astropy.coordinates import SkyCoord

from sunpy.coordinates import propagate_with_solar_surface
from sunpy.coordinates.screens import SphericalScreen
from sunpy.map import Map, pixelate_coord_path, sample_at_coords
from sunpy.net import Fido
from sunpy.net import attrs as a

###############################################################################
# First, we will need to acquire a sequence of images from SDO/AIA.
# We will use a data from 2012 containing a nice example of loop oscillations.

# To keep the online build and download low using a short time range expand to
# 18:05 - 18:30 to see more of the oscillation.

query = Fido.search(
    a.Time('2012-10-20 18:14:00', '2012-10-20 18:19:00'),
    a.Instrument.aia,
    a.Wavelength(171*u.angstrom),
)
files = Fido.fetch(query, site="NSO")
files = sorted(files)


###############################################################################
# Our target will be a set of loops in the corona above an active region.
# First load the FITS files we downloaded above in to map sequence and create a
# rectangular submap or cutout around the area of interest. We will also define
# the path, in this case a line, along which we want to make the time-distance
# plot.

aia_seq = Map(files)
corner = SkyCoord(Tx=-1150*u.arcsec, Ty=-500*u.arcsec,
                  frame=aia_seq[0].coordinate_frame)
ref_sub_map = aia_seq[0].submap(corner, width=250*u.arcsec, height=450*u.arcsec)

line_coords = SkyCoord([-1030, -1057]*u.arcsec, [-220, -206]*u.arcsec,
                       frame=aia_seq[0].coordinate_frame)

###############################################################################
# Next we can plot the full disk map, over-plotting the region of interest and
# also plot the submap and the line along which we want to make the
# time-distance plot.

fig = plt.figure(figsize=(8, 5))
full_disk_ax = fig.add_subplot(121, projection=aia_seq[0])
aia_seq[0].plot(axes=full_disk_ax, clip_interval=[1,99]*u.percent)
aia_seq[0].draw_quadrangle(corner, width=250*u.arcsec, height=450*u.arcsec,
                           axes=full_disk_ax)
sub_map_ax = fig.add_subplot(122, projection=ref_sub_map)
ref_sub_map.plot(axes=sub_map_ax, clip_interval=[1,99]*u.percent)
sub_map_ax.plot_coord(line_coords)

###############################################################################
# There are two approaches that can be used to extract time distance
# measurements from the data:
#
# 1. Reproject all the maps to a common world coordinate system (WCS)
# 2. Transform the coordinates of the line extracted from the reference map
#    to the coordinate systems of the subsequent maps
#
# We will use both and show they achieve almost the same results choosing which
# approach to use will depend on the exact use case.
#
# We will start of with the first approach and reproject all the maps to common WCS
# of the ``ref_sub_map.wcs`` while also taking account of differential rotation of
# using `~sunpy.coordinates.propagate_with_solar_surface` and off-disk pixels using
# `~sunpy.coordinates.screens.SphericalScreen`.

reprojected_sub_maps = []
for cur_map in aia_seq:
    cur_map = cur_map/cur_map.exposure_time
    with (propagate_with_solar_surface(),
          SphericalScreen(cur_map.observer_coordinate, only_off_disk=True)):
        reprojected_sub_maps.append(cur_map.reproject_to(ref_sub_map.wcs, preserve_date_obs=True))
reprojected_sub_maps = Map(reprojected_sub_maps, sequence=True)

###############################################################################
# Now that we have reprojected all the maps to common WCS we can extract the
# pixel coordinates once using :func:`~sunpy.map.pixelate_coord_path` to
# determine the coordinates for every pixel that intersects with the physical
# path and then use :func:`~sunpy.map.sample_at_coords` sample the data at
# these coordinates.

# As the maps are all aligned only need to extract the coordinates once
intensity_coords = pixelate_coord_path(aia_seq[0], line_coords)

intensities_reproject = []

for aia_map in reprojected_sub_maps:
    intensities_reproject.append(sample_at_coords(aia_map, intensity_coords).value)

###############################################################################
# For the second approach we need to transform the ``intensity_coords`` into
# the coordinate frame of each map and then extract the data at corresponding
# pixel coordinates.

intensities_transform = []
for cur_map in aia_seq:
    with propagate_with_solar_surface(), SphericalScreen(cur_map.observer_coordinate,
                                                         only_off_disk=True):
        # The coordinate will automatically be transformed into the cur_map frame
        intensities_transform.append(
            (sample_at_coords(cur_map, intensity_coords)/cur_map.exposure_time).value)

###############################################################################
# Now we have obtained the raw data we need to prepare it for platting and
# calculate the extents of the x and y acies for the final plot.

# The above will give us a list of 1D arrays, one for each map in the sequence.
# We need to stack them into a 2D array.
intensities_reproject = np.vstack(intensities_reproject)
intensities_transform = np.vstack(intensities_transform)

# This defines the distance along the path in arcseconds.
angular_separation = intensity_coords.separation(intensity_coords[0]).to(u.arcsec)

# Get correct values for the extent
extent = [reprojected_sub_maps[0].date.datetime, reprojected_sub_maps[-1].date.datetime,
          0, angular_separation[-1].value]

###############################################################################
# Plot the reference submap, line and extracted data from both methods and
# the difference between them.

fig = plt.figure(figsize=(10, 5), layout="constrained")
left, right = fig.subfigures(nrows=1, ncols=2, width_ratios=[0.6, 0.75])
left_ax = left.add_subplot(111, projection=reprojected_sub_maps[0])
right_ax = right.subplot_mosaic([['repro'], ['trans'], ['diff']],
                                sharex=True, sharey=True)

imag_ax = reprojected_sub_maps[0].plot(axes=left_ax, clip_interval=[1, 99]*u.percent)
plt.colorbar(imag_ax, ax=left_ax)
left_ax.plot_coord(line_coords)

right_ax['repro'].imshow(
    intensities_reproject.T, aspect='auto', interpolation='none',
    extent=extent, cmap=imag_ax.get_cmap()
)
plt.colorbar(right_ax['repro'].images[0], ax=right_ax['repro'])
right_ax['trans'].imshow(
    intensities_transform.T, aspect='auto', interpolation='none',
    extent=extent, cmap=imag_ax.get_cmap()
)
plt.colorbar(right_ax['trans'].images[0], ax=right_ax['trans'])
right_ax['diff'].imshow(
    (intensities_reproject-intensities_transform).T, interpolation='none',
    aspect='auto', extent=extent, cmap='bwr'
)
plt.colorbar(right_ax['diff'].images[0], ax=right_ax['diff'])

locator = mdates.AutoDateLocator(minticks=4)
formatter = mdates.ConciseDateFormatter(locator)
right_ax['diff'].xaxis.set_major_locator(locator)
right_ax['diff'].xaxis.set_major_formatter(formatter)
right_ax['diff'].xaxis.set_minor_locator(mdates.MinuteLocator())

right_ax['repro'].set_title('Reproject')
right_ax['trans'].set_title('Transform')
right_ax['diff'].set_title('Difference')

right_ax['diff'].set_xlabel('Time [UTC]')
right_ax['trans'].set_ylabel('Distance [arcsec]')

plt.show()



"""Tests for EUI Solar Orbiter Map"""

import pytest

import astropy.units as u
from astropy.coordinates import SkyCoord

from sunpy.data.test import get_dummy_map_from_header, get_test_filepath
from sunpy.map.sources import EUIMap
from .helpers import _test_private_date_setters

header_list = [
    get_test_filepath('solo_L1_eui-fsi304-image_20201021T145510206_V03.header'),
]


@pytest.fixture(scope="module", params=header_list)
def eui_map(request):
    return get_dummy_map_from_header(request.param)


def test_EUIMap(eui_map):
    assert isinstance(eui_map, EUIMap)


def test_reference_date(eui_map):
    assert eui_map.reference_date.isot == "2020-10-21T14:55:13.206"


def test_date(eui_map):
    assert eui_map.date.isot == "2020-10-21T14:55:10.206"


def test_private_date_setters(eui_map):
    _test_private_date_setters(eui_map)


def test_is_datasource_for(eui_map):
    assert eui_map.is_datasource_for(eui_map.data, eui_map.meta)


def test_observer_coordinate(eui_map):
    obs_coord = eui_map.observer_coordinate
    assert isinstance(obs_coord, SkyCoord)
    assert obs_coord.obstime.isot == eui_map.meta['date-avg']


def test_observatory(eui_map):
    assert eui_map.observatory == "Solar Orbiter"


def test_measurement(eui_map):
    assert eui_map.measurement == u.Quantity(304, 'angstrom')


def test_exposure_time(eui_map):
    assert eui_map.exposure_time == u.Quantity(6, 's')


def test_level_number(eui_map):
    assert eui_map.processing_level == 1


def test_unit(eui_map):
    assert eui_map.unit == u.Unit('DN')


def test_norm_clip(eui_map):
    # Tests that the default normalizer has clipping disabled
    assert not eui_map.plot_settings['norm'].clip


def test_wcs(eui_map):
    # Smoke test that WCS is valid and can transform from pixels to world coordinates
    eui_map.pixel_to_world(0*u.pix, 0*u.pix)



from sunpy.data.data_manager.downloader import DownloaderBase

# This is the hash of file containing just the character 'a'
MOCK_HASH = "ca978112ca1bbdcafac231b39a23dc4da786eff8147c4e72b9807785afee48bb"


class MockDownloader(DownloaderBase):
    """
    MockDownloader.
    """

    def __init__(self):
        self.times_called = 0
        self.last_called_url = ''

    def download(self, url, path,**kwargs):
        write_to_test_file(path, "a")
        self.times_called += 1
        self.last_called_url = url
        return path


def write_to_test_file(path, contents):
    with open(path, 'w') as f:
        f.write(contents)



import sunpy.timeseries
from sunpy.data.test import get_test_filepath
from sunpy.tests.helpers import figure_test

norh_filepath = get_test_filepath('tca110810_truncated')


def test_implicit_norh():
    # Test a NoRH TimeSeries
    ts_norh = sunpy.timeseries.TimeSeries(norh_filepath)
    assert isinstance(ts_norh, sunpy.timeseries.sources.norh.NoRHTimeSeries)


def test_norh():
    # Test a NoRH TimeSeries
    ts_norh = sunpy.timeseries.TimeSeries(norh_filepath, source='NoRH')
    assert isinstance(ts_norh, sunpy.timeseries.sources.norh.NoRHTimeSeries)


def test_norh_plot_column(norh_test_ts):
    ax = norh_test_ts.plot(columns=['Correlation Coefficient'])
    assert len(ax.lines) == 1
    assert '17GHZ' == ax.lines[0].get_label()


@figure_test
def test_norh_peek(norh_test_ts):
    norh_test_ts.peek()



from unittest import mock

import pytest
from hypothesis import given

import astropy.units as u

import sunpy.net.dataretriever.sources.norh as norh
from sunpy.net import Fido
from sunpy.net import attrs as a
from sunpy.net.fido_factory import UnifiedResponse
from sunpy.net.tests.helpers import mock_query_object
from sunpy.net.tests.strategies import time_attr


@pytest.fixture
def LCClient():
    return norh.NoRHClient()

@pytest.mark.parametrize(("timerange", "url_start", "url_end"), [
    (a.Time('2012/4/21', '2012/4/21'),
     'https://solar.nro.nao.ac.jp/norh/data/tcx/2012/04/tca120421',
     'https://solar.nro.nao.ac.jp/norh/data/tcx/2012/04/tca120421'
     ),
    (a.Time('2012/12/1', '2012/12/2'),
     'https://solar.nro.nao.ac.jp/norh/data/tcx/2012/12/tca121201',
     'https://solar.nro.nao.ac.jp/norh/data/tcx/2012/12/tca121202'
     ),
    (a.Time('2012/3/7', '2012/3/14'),
     'https://solar.nro.nao.ac.jp/norh/data/tcx/2012/03/tca120307',
     'https://solar.nro.nao.ac.jp/norh/data/tcx/2012/03/tca120314'
     )
])
def test_get_url_for_time_range(LCClient, timerange, url_start, url_end):
    with mock.patch('sunpy.net.dataretriever.sources.norh.NoRHClient.search',
                    return_value=mock_query_object(timerange, LCClient)):
        qresponse = LCClient.search(timerange, a.Wavelength(17*u.GHz))
        urls = [i['url'] for i in qresponse]
        assert urls[0] == url_start


@given(time_attr())
def test_can_handle_query(time):
    LCClient = norh.NoRHClient()
    ans1 = LCClient._can_handle_query(time, a.Instrument.norh)
    assert ans1 is True
    ans1 = LCClient._can_handle_query(time, a.Instrument.norh,
                                      a.Wavelength(10*u.GHz))
    assert ans1 is True
    ans2 = LCClient._can_handle_query(time)
    assert ans2 is False


@pytest.mark.remote_data
@pytest.mark.parametrize(("time", "wave"), [(a.Time('2007/08/13', '2007/08/14'),a.Wavelength(17*u.GHz)),(a.Time('2007/08/13', '2007/08/14'), a.Wavelength(34*u.GHz))])
def test_query(time, wave):
    LCClient = norh.NoRHClient()
    qr1 = LCClient.search(time, a.Instrument.norh, wave)
    assert qr1[0]['Start Time'].strftime('%Y-%m-%d') == time.start.strftime('%Y-%m-%d')
    assert qr1[-1]['End Time'].strftime('%Y-%m-%d') == time.end.strftime('%Y-%m-%d')


@pytest.mark.parametrize(("time", "instrument", "wave"), [
    (a.Time('2012/10/4', '2012/10/4'), a.Instrument.norh, a.Wavelength(17*u.GHz)),
    (a.Time('2012/10/4', '2012/10/4'), a.Instrument.norh, a.Wavelength(34*u.GHz))])
def test_get(LCClient, time, instrument, wave):
    with mock.patch('sunpy.net.dataretriever.sources.norh.NoRHClient.search',
                    return_value=mock_query_object(time, LCClient)):
        qr1 = LCClient.search(time, instrument, wave)
        with mock.patch('sunpy.net.dataretriever.sources.norh.NoRHClient.fetch',
                        return_value=mock_query_object(time, LCClient)):
            download_list = LCClient.fetch(qr1)
            assert len(download_list) == len(qr1)


@pytest.mark.parametrize(
    ("time", "instrument", "wave"),
    [(a.Time('2012/10/4', '2012/10/4'), a.Instrument.norh, a.Wavelength(17*u.GHz) | a.Wavelength(34*u.GHz))])
def test_fido(tmp_path, time, instrument, wave):
    with mock.patch('sunpy.net.Fido.search',
                    return_value=UnifiedResponse(mock_query_object(time, LCClient))):
        path = tmp_path / "sub"
        qr = Fido.search(time, instrument, wave)
        assert isinstance(qr, UnifiedResponse)
        with mock.patch('sunpy.net.Fido.fetch',
                        return_value=UnifiedResponse(mock_query_object(time, LCClient))):
            response = Fido.fetch(qr, path=path)
            assert len(response) == len(qr)


def test_attr_reg():
    assert a.Instrument.norh == a.Instrument('NORH')


def test_client_repr(LCClient):
    """
    Repr check
    """
    output = str(LCClient)
    assert output[:50] == 'sunpy.net.dataretriever.sources.norh.NoRHClient\n\nP'


def test_show():
    mock_qr = mock_query_object(a.Time('2016/1/1', '2016/1/1 23:59:59'), norh.NoRHClient())
    qrshow0 = mock_qr.show()
    qrshow1 = mock_qr.show('Wavelength', 'Instrument')
    allcols = {'Start Time', 'End Time', 'Instrument', 'Source',
               'Provider', 'Wavelength', 'url'}
    assert not allcols.difference(qrshow0.colnames)
    assert qrshow1.colnames == ['Wavelength', 'Instrument']
    assert qrshow0['Instrument'][0] == 'NORH'
    assert qrshow1['Wavelength'][0] == 17*u.GHz



import re
import json
import warnings

import numpy as np
from packaging.version import Version

import astropy
from astropy import units as u
from astropy.coordinates import ICRS, SkyCoord
from astropy.table import Column, MaskedColumn
from astropy.utils.data import get_pkg_data_filename
from astropy.utils.masked import Masked

import sunpy.coordinates
from sunpy import log
from sunpy.extern import parse
from sunpy.time import parse_time

__all__ = [
    '_freeze',
    '_map_columns_to_times',
    '_map_columns_to_quantities',
    '_map_event_coord_columns_to_coordinates',
    '_map_chain_code_columns_to_coordinates',
]


def _freeze(obj):
    """
    Create a hashable representation of a dict or list.
    Otherwise returns the original input.
    """
    if isinstance(obj, dict):
        return tuple((k, _freeze(v)) for k, v in obj.items())
    if isinstance(obj, list):
        return tuple(_freeze(elem) for elem in obj)
    return obj


def _get_unit_attributes():
    """
    Returns the attributes from the HEK that are used to parse the values in the HEK table.
    """
    unit_properties_filename = get_pkg_data_filename('data/unit_properties.json',
                                                     package='sunpy.net.hek')
    with open(unit_properties_filename) as unit_file:
        unit_properties = json.load(unit_file)
    return unit_properties["attributes"]


def _get_coord_attributes():
    coord_properties_filename = get_pkg_data_filename('data/coord_properties.json',
                                                      package='sunpy.net.hek')
    with open(coord_properties_filename) as coord_file:
        coord_properties = json.load(coord_file)
    return coord_properties["attributes"]


def _map_columns_to_times(table):
    """
    Represent time columns in an HEK response as astropy Time objects.
    """
    # All time columns from https://www.lmsal.com/hek/VOEvent_Spec.html
    time_keys = ['event_endtime', 'event_starttime', 'event_peaktime']
    for tkey in time_keys:
        if tkey in table.colnames:
            table[tkey] = [
                (parse_time(time, format='iso') if time else np.ma.masked) for time in table[tkey]
            ]


def _map_columns_to_quantities(table):
    """
    For columns in an HEK response which have associated units, combine
    data and unit columns into masked columns with units.
    """
    attributes = _get_unit_attributes() + _get_coord_attributes()
    missing_values = {
        'integer': -999,
        'long': -999,
        'float': np.nan,
    }
    dtype_aliases = {
        'string': 'str',
        'integer': 'int',
    }
    for attr in attributes:
        if (name := attr["name"]) not in table.colnames:
            continue
        if attr.get("is_unit_prop", False):
            continue
        if attr.get("is_chaincode", False):
            continue
        mask = np.array([r is None for r in table[name]])
        # NOTE: Fill with missing values explicitly because None cannot be cast to all dtypes.
        data = np.where(mask, missing_values.get(attr['type'], None), table[name])
        if unit_prop := attr.get("unit_prop", False):
            if unit_prop not in table.colnames:
                log.debug(f"Missing unit property {unit_prop} for {name}. Using event_coordunit.")
                unit_prop = "event_coordunit"
            units = np.array([_parse_unit(u_) if u_ is not None else '' for u_ in table[unit_prop]])
            default_unit = u.dimensionless_unscaled if mask.all() else units[~mask][0]
            units = np.where(mask, default_unit, units).tolist()
            # NOTE: This is done per entry because each entry could, in principle,
            # have a different (though compatible) unit.
            data = u.Quantity([d*_u for d, _u in zip(data, units)])
        dtype = dtype_aliases.get(attr['type'], attr['type'])
        table[name] = MaskedColumn(data=data, mask=mask, name=name, dtype=dtype)
    table.remove_columns(
        [attr["name"] for attr in attributes
         if attr.get("is_unit_prop", False) and attr["name"] in table.colnames]
    )


def _map_event_coord_columns_to_coordinates(table):
    """
    For columns in an HEK response which represent the event coordinates, combine
    these columns into a single column with a SkyCoord object.

    These columns are a special case because the event coordinate information is spread across
    many columns and as such multiple columns have to be used to create a single new
    coordinate column.
    """
    # TODO: Confirm that this is the right thing to do. Previous iterations of this
    # code set the frame to ICRS if the event_coordunit was anything but arcsec.
    # See https://www.lmsal.com/hek/VOEvent_Spec.html for more information.
    frame_mapping = {
        "UTC-HGS-TOPO": sunpy.coordinates.HeliographicStonyhurst,
        "UTC-HPC-TOPO": sunpy.coordinates.Helioprojective,
        "UTC-HGC-TOPO": sunpy.coordinates.HeliographicCarrington,
        "UTC-HCR-TOPO": sunpy.coordinates.Heliocentric,
        "UTC-HRC-TOPO": sunpy.coordinates.Heliocentric,  # Possibly a misspelling of HCR?
    }
    event_coords = []
    for row in table:
        data = [row["event_coord1"], row["event_coord2"]]
        if row["event_coord3"] is not None:
            data.append(row["event_coord3"])
        # NOTE: "event_coordunit" can be space or comma-separated string representing the different
        # units of the different coordinate columns or just a single unit string.
        coord_unit = [_parse_unit(unit_string) for unit_string in re.split(r',\s*|\s+', row["event_coordunit"])]
        if len(coord_unit) == 1:
            coord_unit = len(data) * coord_unit
        data = [d*_u for d, _u in zip(data, coord_unit)]
        obstime = row["event_starttime"]
        frame_type = frame_mapping[row["event_coordsys"]]
        frame_kwargs = {"obstime": obstime}
        if frame_type.name != "heliographic_stonyhurst":
            frame_kwargs["observer"] = sunpy.coordinates.get_earth(obstime)
        frame = frame_type(**frame_kwargs)
        representation_type = "spherical"
        # TODO: Confirm whether this is the right way to represent heliocentric radial coordinates or whether
        # this is even what "heliocentric" refers to.
        if isinstance(frame, sunpy.coordinates.Heliocentric):
            representation_type = "cylindrical"
            # NOTE: See entry for Heliocentric Radial in this table:
            # https://docs.sunpy.org/en/stable/reference/coordinates/index.html#supported-coordinate-systems
            data[0] += 90*u.deg
            # The HCC frame expects the data in the reverse order that the HEK returns them.
            data = data[::-1]
            # NOTE: There seem to be cases where event_coord3 is missing for the case of a Heliocentric frame
            if len(data) == 2:
                data.append(1*u.R_sun)
        event_coord = SkyCoord(*data, frame=frame, representation_type=representation_type)
        event_coords.append(event_coord)
    table.add_column(Column(data=event_coords, name="event_coord"))
    # NOTE: Explicitly not removing event_coordunit because it is used as a unit when parsing
    # the other coordinate columns.
    table.remove_columns([
        'event_coordsys',
        'event_coord1',
        'event_coord2',
        'event_coord3',
    ])


def _map_chain_code_columns_to_coordinates(table):
    """
    Convert columns which contain an HEK "chain code", representing either
    a point or a region, into SkyCoord objects.

    In the case of a point, the whole column can be represented as a single SkyCoord
    object with a time varying frame. In the case of a region, the column can be
    represented as a single SkyCoord if each region has the same shape. This is true
    for the bounding box ("bbox") columns. If the shapes are not equal, the column has
    to be expressed as a list of SkyCoord objects.
    """
    # FIXME: There has to be a better way to do this
    frame_class_mapping = {
        'helioprojective': sunpy.coordinates.Helioprojective,
        'heliographic_stonyhurst': sunpy.coordinates.HeliographicStonyhurst,
        'heliocentric': sunpy.coordinates.Heliocentric,
        'heliographic_carrington': sunpy.coordinates.HeliographicCarrington,
        'icrs': ICRS,
    }
    frame_unit_mapping = {
        'helioprojective': [u.arcsec, u.arcsec],
        'heliographic_stonyhurst': [u.deg, u.deg],
        'heliocentric': [u.R_sun, u.deg],
        'heliographic_carrington': [u.deg, u.deg],
    }
    for attr in _get_coord_attributes():
        if not attr.get('is_chaincode', False):
            continue
        if (name := attr["name"]) not in table.colnames:
            continue
        is_point = attr.get('is_point', False)
        frame_kwargs = {}
        if attr['frame'] != 'icrs':
            frame_kwargs['obstime'] = table['event_starttime']
        if attr['frame'] not in ['heliographic_stonyhurst', 'icrs']:
            frame_kwargs['observer'] = 'earth'
        if attr['frame'] == 'heliocentric':
            frame_kwargs['representation_type'] = 'cylindrical'
        frame = frame_class_mapping[attr['frame']](**frame_kwargs)
        coord_data = []
        shape = (1, 2)  # Only used if all entries are masked
        for row in table[name]:
            if row == '' or row is None:
                data = None
            elif is_point:
                data = np.array(parse('POINT({})', row)[0].split(), dtype=float)
                shape = data.shape
            else:
                data = parse('POLYGON(({}))', row)[0]
                data = np.array([r.split() for r in data.split(',')], dtype=float)
                shape = data.shape
            coord_data.append(data)
        # NOTE: Units are cast to arrays of shape (n_rows,2) to make broadcasting
        # to data arrays easier.
        if attr['frame'] in frame_unit_mapping:
            units = np.array([frame_unit_mapping[attr['frame']]]*len(coord_data))
        else:
            units = np.array(list(map(_parse_unit, table['event_coordunit'])))
            units = np.repeat(units[:, np.newaxis], 2, axis=1)
        # NOTE: Filling in masked values with data of appropriate shape allows for
        # broadcasting of coordinate frame information later on if all shapes are the
        # same.
        coord_data = [np.full(shape, np.nan) if c is None else c for c in coord_data]
        if all([coord_data[0].shape==c.shape for c in coord_data]):
            # NOTE: Taking the transpose of the coordinate such that the first
            # dimension of the coordinate is the number of rows in the table. The
            # transpose is not taken of the data prior to construction so that the
            # frames are appropriately broadcast to the data.
            coordinates = _build_masked_coordinate(coord_data, frame, units).T
        else:
            # NOTE: If the shape of the data in each row is not equal, the column has
            # to be expressed as a list of SkyCoord objects and will have a dtype of
            # object.
            if not frame.shape:
                frame = len(coord_data) * [frame]
            coordinates = [_build_masked_coordinate(c, frame[i], units[i]) for i,c in enumerate(coord_data)]
            # FIXME: The mask is created this way for astropy versions below v7 because SkyCoords cannot
            # do not have a mask attribute until v7 and above. Once our minimum version of astropy is v7
            # this conditional can be removed
            if Version(astropy.__version__) < Version('7'):
                mask = [np.isnan(c.cartesian.xyz).all() for c in coordinates]
            else:
                mask = [c.mask.all() for c in coordinates]
            # NOTE: This is expressed explicitly as a column to avoid stacking the coordinates along the
            # dimension that corresponds to the chaincode.
            coordinates = MaskedColumn(data=coordinates,
                                       name=name,
                                       mask=mask,
                                       dtype='object')
        table[name] = coordinates


def _build_masked_coordinate(data, frame, unit):
    # NOTE: Take the transpose so that entries in the list correspond to
    # components of the coordinate.
    coord_data = []
    for _d, _u in zip(np.array(data).T, unit.T):
        # NOTE: This complexity is to allow for broadcasting of units in cases where
        # there is a single unit or an array of units against a data array that may
        # be multidimensional.
        data = np.stack((_d * _u).ravel()).reshape(_d.shape)
        # FIXME: This conditional is because SkyCoords with masked data are only
        # support in astropy v7 and above. Once our minimum version of astropy
        # is v7, this can be removed.
        if Version(astropy.__version__) < Version('7'):
            coord_data.append(data)
        else:
            coord_data.append(Masked(data, mask=np.isnan(_d)))
    if frame.name == 'heliocentric':
        # There is a 90 degree offset between Heliocentric radial and the
        # cylindrical representation of the Heliocentric as defined in sunpy.
        # See https://docs.sunpy.org/en/stable/reference/coordinates/index.html#supported-coordinate-systems.
        coord_data[-1] += 90*u.deg
        # The HEK only returns two coordinates, but a Heliocentric frame requires three.
        # Here, we assume that z=1 R_sun.
        coord_data += [1*u.R_sun]
    return SkyCoord(*coord_data, frame=frame)


def _parse_unit(unit_string):
    """
    Parses an HEK string representation of a unit and converts it into a `astropy.units.Unit`.

    Parameters
    ----------
    unit_string : `str`
        String representation of unit as return by the HEK.

    Returns
    -------
    unit : `~astropy.units.Unit`

    Notes
    ----
    `A complete list of HEK parameters.
    <https://www.lmsal.com/hek/VOEvent_Spec.html>`__
    """
    # NOTE: This has to be done as a regular expression because "ma" is already
    # an astropy unit representing "milliannum" (1/1000 years) and as such it
    # cannot be used as an alias for "milliampere".
    unit_string = re.sub(r"\bma\b", "mA", unit_string)
    # NOTE: These are done separately because unit names with spaces cannot simply be aliased
    # and thus have to be defined as new units. Units without spaces that have unconventional
    # spellings can simply be aliased.
    aliases = {
        "arcseconds": u.arcsec,
        "degrees": u.deg,
        "sec": u.s,
        "emx": u.Mx,
        "amperes": u.A,
        "ergs": u.erg,
        "radians": u.rad,
        "hz": u.Hertz,
        "rsun": u.R_sun,
    }
    enabled_units = [
        u.def_unit("cubic centimeter", represents=u.cm**3),
        u.def_unit("square centimeter", represents=u.cm**2),
        u.def_unit("cubic meter", represents=u.m**3),
        u.def_unit("square meter", represents=u.m**2),
        u.def_unit("ergs per cubic centimeter", represents=u.erg/u.cm**3),
        u.def_unit("ergs/cm^3", represents=u.erg/u.cm**3),
        u.def_unit("HMI pixels", represents=0.5*u.arcsec/u.pixel),
        u.def_unit("HMI pixel", represents=0.5*u.arcsec/u.pixel),
    ]
    # NOTE: This logic accounts for the scenario where multiple units may be contained in
    # a single string. However, this logic should not be applied to custom units which contain
    # spaces, e.g. those custom units enabled above.
    if unit_string not in aliases and unit_string not in [eu.name for eu in enabled_units]:
        split_unit_string = re.split(r',\s*|\s+', unit_string)
        if len(split_unit_string) > 1:
            log.debug(f"Multiple units found in {unit_string}. Using only the first unit.")
            unit_string = split_unit_string[0]
    with u.add_enabled_units(enabled_units), u.set_enabled_aliases(aliases), warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=u.UnitsWarning, message='.*contains multiple slashes*')
        parsed_unit = u.Unit(unit_string)
        return parsed_unit



from pydot import Dot, Edge, Node

graph = Dot("my_graph", graph_type="graph", bgcolor="white")

core_kwargs = {'style': 'filled', 'fillcolor': '#9bdd99'}
optional_kwargs = {'style': 'filled', 'fillcolor': '#9fc5e8'}
external_kwargs = {'style': 'filled', 'fillcolor': '#efa7a7'}

# Core curriculum
graph.add_node(Node('Installing', **core_kwargs))
graph.add_node(Node('Units', **core_kwargs))
graph.add_edge(Edge('Installing', 'Units', dir='forward'))

graph.add_node(Node('Times', **core_kwargs))
graph.add_edge(Edge('Units', 'Times', dir='forward'))

graph.add_node(Node('Coordinates', **core_kwargs))
graph.add_edge(Edge('Times', 'Coordinates', dir='forward'))

graph.add_node(Node('Acquiring data', **core_kwargs))
graph.add_edge(Edge('Coordinates', 'Acquiring data', dir='forward'))

graph.add_node(Node('Maps', **core_kwargs))
graph.add_edge(Edge('Acquiring data', 'Maps', dir='forward'))

graph.add_node(Node('Timeseries', **core_kwargs))
graph.add_edge(Edge('Acquiring data', 'Timeseries', dir='forward'))

# Specific acquiring data tutorials
graph.add_node(Node('Acquiring JSOC data', **optional_kwargs))
graph.add_edge(Edge('Acquiring data', 'Acquiring JSOC data', dir='forward'))

graph.add_node(Node('Searching HEK', **optional_kwargs))
graph.add_edge(Edge('Acquiring data', 'Searching HEK', dir='forward'))

graph.add_node(Node('Working with GOES XRS', **optional_kwargs))
graph.add_edge(Edge('Timeseries', 'Working with GOES XRS', dir='forward'))

# External tutorials
graph.add_node(Node('AIAPy tutorial', **external_kwargs))
graph.add_edge(Edge('Maps', 'AIAPy tutorial', dir='forward'))

graph.add_node(Node('Querying Helioviewer.org', **external_kwargs))
graph.add_edge(Edge('Acquiring data', 'Querying Helioviewer.org', dir='forward'))

graph.write_png("tutorial_structure.png")




import numpy as np

from sunpy.data.test import get_test_filepath
from sunpy.io import _fits, _jp2
from sunpy.io._header import FileHeader
from sunpy.tests.helpers import skip_glymur

AIA_193_JP2 = get_test_filepath("2013_06_24__17_31_30_84__SDO_AIA_AIA_193.jp2")
EUI_174_JP2 = get_test_filepath("2022_04_01__00_00_45__SOLO-EUI-FSI_EUI_FSI_174.jp2")
TEST_AIA_IMAGE = get_test_filepath('aia_171_level1.fits')


@skip_glymur
def test_read():
    data, header = _jp2.read(AIA_193_JP2)[0]
    assert isinstance(data, np.ndarray)
    assert data.shape == (410, 410)
    assert isinstance(header, FileHeader)


@skip_glymur
def test_read_header():
    header = _jp2.get_header(EUI_174_JP2)[0]
    assert isinstance(header, FileHeader)
    # Check that the header has been parsed correctly
    # So we expect some FITS keywords, Keycomments and History
    assert header["KEYCOMMENTS"]['XTENSION'] == "binary table extension"
    # We check the first line to see if the header has been parsed correctly
    assert "orkingDirectory /tmp/telemetry_parser --configFile /home/eui/config/conf" in header['HISTORY']


@skip_glymur
def test_read_memmap():
    data, _ = _jp2.read(AIA_193_JP2, memmap=True)[0]
    # The data is shared, with what, I am unclear
    assert data.base is not None
    # Keyword is not passed in the function call
    data, _ = _jp2.read(AIA_193_JP2, memmap=False)[0]
    assert data.base is not None


@skip_glymur
def test_simple_write(tmpdir):
    data, header = _fits.read(TEST_AIA_IMAGE)[0]
    outfile = tmpdir / "test.jp2"
    _jp2.write(str(outfile), data, header)
    assert outfile.exists()

    # Sanity check that reading back the jp2 returns coherent data
    jp2_readback = _jp2.read(outfile)
    assert header['DATE'] == jp2_readback[0].header['DATE']

    # jp2 requires the data array to have type uint8, so cast the original
    # data array to uint8 to compare it with the generated jp2 file.
    original_data = np.uint8(data)
    assert np.array_equal(original_data, jp2_readback[0].data)



"""
================
Using TimeSeries
================

This example is intended to demonstrate the current state of `~sunpy.timeseries.TimeSeries`.
"""
import datetime
from collections import OrderedDict

import matplotlib.pyplot as plt
import numpy as np
from pandas import DataFrame

import astropy.units as u
from astropy.time import Time, TimeDelta

import sunpy.data.sample
import sunpy.timeseries
from sunpy.net import Fido
from sunpy.net import attrs as a
from sunpy.time import TimeRange, parse_time
from sunpy.util.metadata import MetaDict

##############################################################################
# We can create a range of supported timeseries:
# There is a ``source`` keyword that allows one to specify the source of the data
# It should be auto-detected in most cases.

ts_eve = sunpy.timeseries.TimeSeries(sunpy.data.sample.EVE_TIMESERIES, source='EVE')
ts_goes = sunpy.timeseries.TimeSeries(sunpy.data.sample.GOES_XRS_TIMESERIES, source='XRS')
ts_lyra = sunpy.timeseries.TimeSeries(sunpy.data.sample.LYRA_LEVEL3_TIMESERIES, source='LYRA')
ts_norh = sunpy.timeseries.TimeSeries(sunpy.data.sample.NORH_TIMESERIES, source='NoRH')
ts_rhessi = sunpy.timeseries.TimeSeries(sunpy.data.sample.RHESSI_TIMESERIES, source='RHESSI')
ts_gbm = sunpy.timeseries.TimeSeries(sunpy.data.sample.GBM_TIMESERIES, source='GBMSummary')

##############################################################################
# You can create a list of timeseries when using multiple files.
# First, we shall download these files using `~sunpy.net.Fido`.

goes = Fido.search(a.Time("2012/06/01", "2012/06/04"), a.Instrument.xrs)
goes_files = Fido.fetch(goes)

# Using these new files you get a list of timeseries
list_of_goes_ts = sunpy.timeseries.TimeSeries(goes_files, source='XRS')

##############################################################################
# You can concatenate them together using `~sunpy.timeseries.GenericTimeSeries.concatenate`
# or when creating the `~sunpy.timeseries.TimeSeries`

combined_goes_ts = sunpy.timeseries.TimeSeries(goes_files, source='XRS', concatenate=True)
# Manually
combined_goes_ts = list_of_goes_ts[0].concatenate(list_of_goes_ts[1])

fig, ax = plt.subplots()
combined_goes_ts.plot(axes=ax)

plt.show()

##############################################################################
# The TimeSeries object has 3 primary components:
#
# ``.data`` : The internal data representation. If you want the underlying data,
# use `~sunpy.timeseries.GenericTimeSeries.to_dataframe()`
# ``.meta`` : Stores the metadata that is able to be parsed from the data files
# ``.units`` : Stores the units for each column, with keys that match the name of each column.

# This will give you a useful dataframe to manipulate the data with.
lyra_data = ts_lyra.to_dataframe()
lyra_data

##############################################################################
# This will give you the metadata
ts_lyra.meta

##############################################################################
# This will give you the units
ts_lyra.units

##############################################################################
# There are a couple of other useful properties:

# The time range of the data, the name of the data columns
ts_lyra.time_range, ts_lyra.columns

##############################################################################
# Further data is available from within the metadata, you can filter out for a
# key using the :meth:`.TimeSeriesMetaData.get` method.

combined_goes_ts.meta.get("publisher_name")

##############################################################################
# You can access a specific value within the `~sunpy.timeseries.TimeSeries` data
# using all the normal `pandas` methods.
# For example, to get the row with the index of "2015-01-01 00:02:00.008000"
# Pandas will actually parse a string to a datetime automatically if it can:

lyra_data.loc['2011-06-07 00:02:00.010']
# If this fails, you will need to use parse_time to convert the string to a datetime
lyra_data.loc[parse_time('2011-06-07 00:02:00.010').datetime]
# Pandas includes methods to find the indexes of the max/min values in a dataframe:
ts_lyra.to_dataframe()['CHANNEL1'].idxmax(), ts_lyra.to_dataframe()['CHANNEL1'].idxmin()

##############################################################################
# An individual column can be extracted

ts_eve.extract('CMLon')

##############################################################################
# Changing the units for a column simply requires changing the value

ts_eve.units['a'] = u.m

##############################################################################
# Quantities can be extracted from a column using :meth:`sunpy.timeseries.GenericTimeSeries.quantity`

colname = 'CMLat'
quantity = ts_eve.quantity(colname)
quantity

##############################################################################
# You can add or overwrite a column using :meth:`sunpy.timeseries.GenericTimeSeries.add_column`.
# This method only accepts an `~astropy.units.Quantity` and will convert to the intended units
# if necessary.

new_quantity = quantity.value * 0.01 * ts_eve.units[colname]
new_eve_ts = ts_eve.add_column(colname, new_quantity, overwrite=True)

##############################################################################
# You can truncate using the :meth:`sunpy.timeseries.GenericTimeSeries.truncate` method.

ts_goes_trunc = ts_goes.truncate(0, 100000, 2)
# Or using a `TimeRange`
ts_goes_trunc = ts_goes.truncate(TimeRange('2011-06-07 05:00', '2011-06-07 06:30'))
# Or using strings
ts_goes_trunc = ts_goes.truncate('2011-06-07 05:00', '2011-06-07 06:30')

fig, ax = plt.subplots()
ts_goes_trunc.plot(axes=ax)

plt.show()

##############################################################################
# For now you can only resample by using `pandas`.
# Changing values within the dataframe directly will often affect the units
# involved, but these won't be picked up by `~sunpy.timeseries.TimeSeries`.
# Take care when doing this to ensure dimensional consistency.

df_downsampled = ts_goes_trunc.to_dataframe().resample('10T').mean()
ts_downsampled = sunpy.timeseries.TimeSeries(df_downsampled,
                                             ts_goes_trunc.meta,
                                             ts_goes_trunc.units)

fig, ax = plt.subplots()
ts_downsampled.plot(axes=ax)
plt.show()

##############################################################################
# The data from the `~sunpy.timeseries.TimeSeries` can be retrieved in a number of formats

# pandas DataFrame
ts_goes.to_dataframe()
# astropy Table
ts_goes.to_table()
# numpy array
ts_goes.to_array()

##############################################################################
# Creating a `~sunpy.timeseries.TimeSeries` from scratch can be done several ways.
# Input data can be in the form of a `pandas.DataFrame` (preferred),
# an `astropy.table.Table` or a `numpy.array`.

base = datetime.datetime.today()
dates = Time(base) - TimeDelta(np.arange(24 * 60)*u.minute)
intensity = np.sin(np.arange(0, 12 * np.pi, ((12 * np.pi) / (24 * 60))))
# Create the data DataFrame, header MetaDict and units OrderedDict
data = DataFrame(intensity, index=dates, columns=['intensity'])
meta = MetaDict({'key': 'value'})
units = OrderedDict([('intensity', u.W / u.m**2)])
# Create the TimeSeries
ts_custom = sunpy.timeseries.TimeSeries(data, meta, units)

fig, ax = plt.subplots()
ts_custom.plot(axes=ax)
plt.show()



"""
This module provides fundamental solar physical constants.
"""
import io

from astropy.table import Table
from astropy.time import Time

from sunpy.sun import _constants as _con

__all__ = [
    'get', 'find', 'print_all', 'spectral_classification', 'au', 'mass', 'equatorial_radius',
    'volume', 'surface_area', 'average_density', 'equatorial_surface_gravity',
    'effective_temperature', 'luminosity', 'mass_conversion_rate', 'escape_velocity', 'sfu',
    'average_angular_size', 'sidereal_rotation_rate', 'first_carrington_rotation',
    'mean_synodic_period'
]

constants = _con.physical_constants


def get(key):
    """
    Retrieve a constant by key. This is just a short cut into a dictionary.

    Parameters
    ----------
    key : `str`
        Key in dictionary in ``constants``.

    Returns
    -------
    constant : `~astropy.constants.Constant`

    See Also
    --------
    `~sunpy.sun.constants` :
        Contains the description of ``constants``, which, as a dictionary literal object, does not
        itself possess a docstring.

    Examples
    --------
    >>> from sunpy.sun import constants
    >>> constants.get('mass')
    <<class 'astropy.constants.iau2015.IAU2015'> name='Solar mass' value=1.9884754153381438e+30 uncertainty=9.236140093538353e+25 unit='kg' reference='IAU 2015 Resolution B 3 + CODATA 2014'>
    """
    ret = constants[key]
    ret.__doc__ = ret.name
    return ret


def find(sub=None):
    """
    Return list of constants keys containing a given string.

    Parameters
    ----------
    sub : `str`, optional
        Sub-string to search keys for. By default set to `None` and returns all keys.

    Returns
    -------
    `None`, `list`
        The matching keys.

    See Also
    --------
    `~sunpy.sun.constants` :
        Contains the description of ``constants``, which, as a dictionary literal object, does not itself possess a docstring.
    """
    if sub is None:
        result = list(constants.keys())
    else:
        result = [key for key in constants if sub.lower() in key.lower()]

    result.sort()
    return result


def print_all():
    """
    Provides a table of the complete list of constants.

    Returns
    -------
    `astropy.table.Table`
    """
    data_rows = []
    for key, this_constant in constants.items():
        data_rows.append([
            key, this_constant.name, this_constant.value, this_constant.uncertainty,
            str(this_constant.unit), this_constant.reference
        ])

    t = Table(rows=data_rows, names=('key', 'name', 'value', 'uncertainty', 'unit', 'Reference'))
    return t


def _build_docstring():
    """Build docstring containing RST-formatted table of constants."""
    lines = ['The following constants are available:\n']

    rows = []
    for key, const in constants.items():
        rows.append([key, const.value, const._unit_string, const.name])

    table = Table(rows=rows, names=('Name', 'Value', 'Unit', 'Description'))
    table['Value'].info.format = '14.9g'

    f = io.StringIO()
    table.write(f, format='ascii.rst')
    lines.append(f.getvalue())

    return '\n'.join(lines)


# Add a table of constants to the docs
if __doc__ is not None:
    __doc__ += _build_docstring()

# Spectral class is not included in physical constants since it is not a number
#: Spectral classification
spectral_classification = 'G2V'

au = astronomical_unit = get('mean distance')
# The following variables from _gets are brought out by making them
# accessible through a call such as sun.volume
mass = get('mass')
equatorial_radius = radius = get('radius')
volume = get('volume')
surface_area = get('surface area')
average_density = density = get('average density')
equatorial_surface_gravity = surface_gravity = get('surface gravity')
effective_temperature = get('effective temperature')
luminosity = get('luminosity')
mass_conversion_rate = get('mass conversion rate')
escape_velocity = get('escape velocity')
sfu = get('solar flux unit')
# Observable parameters
average_angular_size = get('average angular size')
sidereal_rotation_rate = get('sidereal rotation rate')

#: Time of the start of the first Carrington rotation
first_carrington_rotation = Time(get('first Carrington rotation (JD TT)'), format='jd', scale='tt')
mean_synodic_period = get('mean synodic period')



"""A MapSequence object"""

import html
import textwrap
import webbrowser
from copy import deepcopy
from tempfile import NamedTemporaryFile

import matplotlib.animation
import numpy as np

from astropy.visualization import ImageNormalize

from sunpy.map import GenericMap
from sunpy.map.maputils import _clip_interval, _handle_norm
from sunpy.util import expand_list
from sunpy.util.exceptions import warn_user
from sunpy.visualization import axis_labels_from_ctype, wcsaxes_compat

__all__ = ['MapSequence']


class MapSequence:
    """
    MapSequence

    A series of Maps in a single object.

    Parameters
    ----------
    args : `list`
        A list of Map instances
    sortby : { "date" | `None`}
        Method by which the MapSequence should be sorted along the z-axis.
        Defaults to sorting by: "date" and is the only supported sorting strategy.
        Passing `None` will disable sorting.

    Attributes
    ----------
    maps : `list`
        This attribute holds the list of Map instances obtained from parameter args.

    Notes
    -----
    To coalign a mapsequence so that solar features remain on the same pixels,
    please see the "Coalignment of MapSequences" note below.

    .. note:: The individual components of a `~sunpy.map.MapSequence` can be coaligned
              using the functions in `sunkit_image.coalignment` or using the
              `~sunpy.map.GenericMap.reproject_to` method.

    Examples
    --------
    >>> import sunpy.map
    >>> mapsequence = sunpy.map.Map('images/*.fits', sequence=True)   # doctest: +SKIP
    """

    def __init__(self, *args, sortby='date'):
        """Creates a new Map instance"""

        self.maps = expand_list(args)

        for m in self.maps:
            if not isinstance(m, GenericMap):
                raise ValueError('MapSequence expects pre-constructed map objects.')

        # Optionally sort data
        if sortby is not None:
            if sortby not in self._sort_methods:
                raise ValueError(f"sortby must be one of the following: {list(self._sort_methods.keys())}")
            self.maps.sort(key=self._sort_methods[sortby])


    @property
    def _sort_methods(self):
        return {
            "date": lambda m: m.date,
        }

    def __getitem__(self, key):
        """Overriding indexing operation. If the key results in a single map,
        then a map object is returned. This allows functions like enumerate to
        work. Otherwise, a mapsequence is returned."""

        if isinstance(self.maps[key], GenericMap):
            return self.maps[key]
        else:
            return MapSequence(self.maps[key])

    def __len__(self):
        """Return the number of maps in a mapsequence."""
        return len(self.maps)

    def __repr__(self):
        names = set([m.__class__.__name__ for m in self.maps])
        return (object.__repr__(self) + "\n" +
                f'MapSequence of {len(self.maps)} elements, with maps from {", ".join(names)}')

    def _repr_html_(self):
        nmaps = len(self)

        # Output a warning about rendering time if there are more than 9 Maps
        if nmaps > 9:
            warn_user(f"Rendering the summary for a MapSequence of {nmaps} Maps "
                      "may take a while.")

        # Assemble the individual HTML repr from each Map, all hidden initially
        repr_list = [f"<div style='display: none' index={i}>{m._repr_html_()}</div>"
                     for i, m in enumerate(self.maps)]

        # Unhide the first Map
        repr_list_html = "\n".join(repr_list).replace('display: none', 'display: ', 1)

        # Return HTML with Javascript-powered buttons
        # To avoid potential conflicts, the Javascript code does not use any user-defined functions
        return textwrap.dedent(f"""\
            <pre>{html.escape(self.__repr__())}</pre>
            <form cur_index=0 max_index={nmaps - 1}>
                <!-- Button to decrement index (always starts disabled) -->
                <input type=button value='&larr;' style='font-weight: bold' disabled onClick='
                    var form = this.parentElement;

                    // Decrement index if allowed
                    var cur_index = Math.max(
                                        parseInt(form.getAttribute("cur_index")) - 1,
                                        0
                                    );
                    form.setAttribute("cur_index", cur_index);

                    // Enable the decrement button if and only if this is not the first Map
                    form.children[0].disabled = (cur_index == 0);

                    // Always enable the increment button (because we just decremented)
                    form.children[1].disabled = false;

                    // Update string (which is children[2] of the form)
                    form.children[2].innerHTML = "Map at index " + cur_index;

                    // Update visibilities to show only the current index
                    // This avoids for...of syntax to retain support for ES5 browsers (e.g., IE11)
                    var array = Array.prototype.slice.call(form.lastElementChild.children);
                    array.forEach(function (elem)
                        {{
                            var form = elem.parentElement.parentElement;
                            elem.style.display = (elem.getAttribute("index") ==
                                                      form.getAttribute("cur_index") ? "" : "none"
                                                 );
                        }}
                    );
                '/>

                <!-- Button to increment index (starts enabled if there is more than one Map) -->
                <input type=button value='&rarr;' style='font-weight: bold'
                    {"" if nmaps > 1 else "disabled"} onClick='

                    var form = this.parentElement;

                    // Increment index if allowed
                    var cur_index = Math.min(
                                        parseInt(form.getAttribute("cur_index")) + 1,
                                        form.getAttribute("max_index")
                                    );
                    form.setAttribute("cur_index", cur_index);

                    // Always enable the decrement button (because we just incremented)
                    form.children[0].disabled = false;

                    // Enable the increment button if and only if this is not the last Map
                    form.children[1].disabled = (cur_index == form.getAttribute("max_index"));

                    // Update string (which is children[2] of the form)
                    form.children[2].innerHTML = "Map at index " + cur_index;

                    // Update visibilities to show only the current index
                    // This avoids for...of syntax to retain support for ES5 browsers (e.g., IE11)
                    var array = Array.prototype.slice.call(form.lastElementChild.children);
                    array.forEach(function (elem)
                        {{
                            var form = elem.parentElement.parentElement;
                            elem.style.display = (elem.getAttribute("index") ==
                                                      form.getAttribute("cur_index") ? "" : "none"
                                                 );
                        }}
                    );

                '/>

                <!-- This string is updated as the index is changed -->
                <span>Map at index 0</span>

                <!-- This element is at the end so that lastElementChild will point to it -->
                <div>
                    {repr_list_html}
                </div>
            </form>""")

    def quicklook(self):
        """
        Display a quicklook summary of the MapSequence instance using the default web browser.

        Click on the |larr| and |rarr| buttons to step through the individual maps.

        .. |larr|   unicode:: U+02190 .. LEFTWARDS ARROW
        .. |rarr|   unicode:: U+02192 .. RIGHTWARDS ARROW

        Notes
        -----
        The image colormap uses
        `histogram equalization <https://en.wikipedia.org/wiki/Histogram_equalization>`__.

        Interactive elements require Javascript support to be enabled in the web browser.

        Examples
        --------
        >>> from sunpy.map import Map
        >>> import sunpy.data.sample  # doctest: +REMOTE_DATA
        >>> seq = Map(sunpy.data.sample.HMI_LOS_IMAGE,
        ...           sunpy.data.sample.AIA_1600_IMAGE,
        ...           sunpy.data.sample.EIT_195_IMAGE,
        ...           sequence=True)  # doctest: +REMOTE_DATA +IGNORE_WARNINGS
        >>> seq.quicklook()  # doctest: +SKIP

        (which will open the following content in the default web browser)

        .. generate:: html
            :html_border:

            from sunpy.map import Map
            import sunpy.data.sample
            seq = Map(sunpy.data.sample.HMI_LOS_IMAGE,
                      sunpy.data.sample.AIA_1600_IMAGE,
                      sunpy.data.sample.EIT_195_IMAGE,
                      sequence=True)
            print(seq._repr_html_())

        """
        with NamedTemporaryFile('w', delete=False, prefix='sunpy.map.', suffix='.html') as f:
            url = 'file://' + f.name
            f.write(textwrap.dedent(f"""\
                <html>
                    <title>Quicklook summary for {html.escape(object.__repr__(self))}</title>
                    <body>{self._repr_html_()}</body>
                </html>"""))
        webbrowser.open_new_tab(url)

    def plot(self, axes=None, annotate=True,
             interval=200, plot_function=None, clip_interval=None, **kwargs):
        """
        A animation plotting routine that animates each element in the
        MapSequence

        Parameters
        ----------
        axes : matplotlib.axes.Axes
            axes to plot the animation on, if none uses current axes
        annotate : bool
            Annotate the figure with scale and titles
        interval : int
            Animation interval in ms
        plot_function : function
            A function to be called as each map is plotted.
            For more information see `sunpy.visualization.animator.MapSequenceAnimator`.
        clip_interval : two-element `~astropy.units.Quantity`, optional
            If provided, the data will be clipped to the percentile interval bounded by the two
            numbers.
        norm : `matplotlib.colors.Normalize` or `str`
            Normalization used in scaling the plot.
            This will override the predefined value in ``plot_settings['norm']``.
        cmap : `matplotlib.colors.Colormap` or `str`
            Color map to be used in coloring the plot.
            This will override the predefined value in ``plot_settings['cmap']``.

        Returns
        -------
        `matplotlib.animation.FuncAnimation`
            A FuncAnimation instance.

        See Also
        --------
        `sunpy.visualization.animator.MapSequenceAnimator`

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> import matplotlib.animation as animation
        >>> from sunpy.map import Map

        >>> sequence = Map(files, sequence=True)   # doctest: +SKIP
        >>> ani = sequence.plot(colorbar=True)   # doctest: +SKIP
        >>> plt.show()   # doctest: +SKIP

        Save an animation of the MapSequence

        >>> sequence = Map(res, sequence=True)   # doctest: +SKIP
        >>> ani = sequence.plot()   # doctest: +SKIP
        >>> Writer = animation.writers['ffmpeg']   # doctest: +SKIP
        >>> writer = Writer(fps=10, metadata=dict(artist='SunPy'), bitrate=1800)   # doctest: +SKIP
        >>> ani.save('mapsequence_animation.mp4', writer=writer)   # doctest: +SKIP

        Save an animation with the limb at each time step

        >>> def myplot(fig, ax, sunpy_map):
        ...    p = sunpy_map.draw_limb()
        ...    return p
        >>> sequence = Map(files, sequence=True)   # doctest: +SKIP
        >>> ani = sequence.peek(plot_function=myplot)   # doctest: +SKIP
        >>> plt.show()   # doctest: +SKIP
        """
        axes = self[0]._check_axes(axes)
        fig = axes.get_figure()

        if not plot_function:
            def plot_function(fig, ax, smap):
                return []
        removes = []

        # Normal plot
        def annotate_frame(i, title=None):
            axes.set_title(f"{self[i].name}" if title is None else title)
            axes.set_xlabel(axis_labels_from_ctype(self[i].coordinate_system[0],
                                                   self[i].spatial_units[0]))
            axes.set_ylabel(axis_labels_from_ctype(self[i].coordinate_system[1],
                                                   self[i].spatial_units[1]))

        ani_data = self.maps
        im = ani_data[0].plot(axes=axes, **kwargs)

        def updatefig(i, im, annotate, ani_data, removes):
            while removes:
                removes.pop(0).remove()
            im.set_array(ani_data[i].data)
            im.set_cmap(kwargs.get('cmap', ani_data[i].plot_settings.get('cmap')) or "grey")
            norm = deepcopy(kwargs.get('norm', ani_data[i].plot_settings.get('norm')))
            if clip_interval is not None:
                vmin, vmax = _clip_interval(ani_data[i].data, clip_interval)
                if norm is None:
                    norm = ImageNormalize()
                norm.vmin=vmin
                norm.vmax=vmax
            if norm:
                _handle_norm(norm, kwargs)
                im.set_norm(norm)

            im.axes.reset_wcs(ani_data[i].wcs)
            wcsaxes_compat.default_wcs_grid(axes)

            if annotate:
                annotate_frame(i, title=kwargs.get('title'))
            removes += list(plot_function(fig, axes, ani_data[i]))

        ani = matplotlib.animation.FuncAnimation(fig, updatefig,
                                                 frames=list(range(0, len(ani_data))),
                                                 fargs=[im, annotate, ani_data, removes],
                                                 interval=interval,
                                                 blit=False)

        return ani

    def peek(self, **kwargs):
        """
        A animation plotting routine that animates each element in the
        MapSequence

        Parameters
        ----------
        fig : matplotlib.figure.Figure
            Figure to use to create the explorer
        annotate : bool
            Annotate the figure with scale and titles
        interval : int
            Animation interval in ms
        colorbar : bool
            Plot colorbar
        plot_function : function
            A function to call to overplot extra items on the map plot.
            For more information see `sunpy.visualization.animator.MapSequenceAnimator`.

        Returns
        -------
        mapsequenceanim : `sunpy.visualization.animator.MapSequenceAnimator`

        See Also
        --------
        sunpy.visualization.animator.MapSequenceAnimator

        Examples
        --------
        >>> import matplotlib.pyplot as plt
        >>> from sunpy.map import Map

        >>> sequence = Map(files, sequence=True)   # doctest: +SKIP
        >>> ani = sequence.peek(colorbar=True)   # doctest: +SKIP
        >>> plt.show()   # doctest: +SKIP

        Plot the map with the limb at each time step

        >>> def myplot(fig, ax, sunpy_map):
        ...    p = sunpy_map.draw_limb()
        ...    return p
        >>> sequence = Map(files, sequence=True)   # doctest: +SKIP
        >>> ani = sequence.peek(plot_function=myplot)   # doctest: +SKIP
        >>> plt.show()   # doctest: +SKIP

        Decide you want an animation:

        >>> sequence = Map(files, sequence=True)   # doctest: +SKIP
        >>> ani = sequence.peek(colorbar=True)   # doctest: +SKIP
        >>> mplani = ani.get_animation()   # doctest: +SKIP
        """
        # Move the import for speed reasons
        from sunpy.visualization.animator.mapsequenceanimator import MapSequenceAnimator

        return MapSequenceAnimator(self, **kwargs)

    @property
    def all_same_shape(self):
        """
        True if the data array of each map has the same shape.
        """
        return np.all([m.data.shape == self.maps[0].data.shape for m in self.maps])


    @property
    def data(self):
        """
        Data array of shape ``(N_y, N_x, N_t)`` where ``(N_y,N_x)`` is the
        shape of each individual map and ``N_t`` is the number of maps.

        .. note:: If all maps do not have the same shape, a `ValueError` is raised.
        """
        if not self.all_same_shape:
            raise ValueError('Not all maps have the same shape.')
        data = np.asarray([m.data for m in self.maps])
        return np.swapaxes(np.swapaxes(data, 0, 1), 1, 2)

    @property
    def mask(self):
        """
        Combined mask with same shape as `~sunpy.map.MapSequence.data` for all maps in the sequence.

        If no map in the sequence has a mask, this returns None.
        If at least one map in the sequence has a mask, the layers
        corresponding to those maps without a mask will be all `False`.
        """
        if not np.any([m.mask is not None for m in self.maps]):
            return None
        mask = np.zeros_like(self.data, dtype=bool)
        for i, m in enumerate(self):
            if m.mask is not None:
                mask[..., i] = m.mask
        return mask

    @property
    def meta(self):
        """
        Metadata objects of each map as a list.
        """
        return [m.meta for m in self.maps]

    def save(self, filepath, filetype='auto', **kwargs):
        """
        Saves the sequence as one map for FITS file.

        Parameters
        ----------
        filepath : `str`
            Template string specifying the file to which each map is saved.
            The string must contain ``"{index}"``, which will be populated with
            the corresponding index number for each map. Format specifiers
            (e.g., ``"{index:03}"``) can be used.
        filetype : `str`
            'auto' or any supported file extension.
        kwargs :
            Any additional keyword arguments are passed to
            `~sunpy.map.GenericMap.save`.

        Examples
        --------
        >>> from sunpy.map import Map
        >>> import sunpy.data.sample # doctest: +SKIP
        >>> smap = Map(sunpy.data.sample.HMI_LOS_IMAGE,
        ...            sunpy.data.sample.AIA_1600_IMAGE,
        ...            sequence=True)  # doctest: +SKIP
        >>> smap.save('map_{index:03}.fits')  # doctest: +SKIP
        """
        if filepath.format(index=0) == filepath:
            raise ValueError("'{index}' must be appear in the string")

        for index, map_seq in enumerate(self.maps):
            map_seq.save(filepath.format(index=index), filetype, **kwargs)
